# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from logging import Logger, getLogger
from typing import Any, Callable, Optional, TypeVar, Union

from anyio import lowlevel
from asyncer import create_task_group

from ....doc_utils import export_module
from ....llm_config import LLMConfig
from ....tools import Tool
from .clients.realtime_client import RealtimeClientProtocol, get_client
from .function_observer import FunctionObserver
from .realtime_observer import RealtimeObserver

F = TypeVar("F", bound=Callable[..., Any])

global_logger = getLogger(__name__)


@dataclass
class RealtimeAgentCallbacks:
    """Callbacks for the Realtime Agent."""

    # async empty placeholder function
    on_observers_ready: Callable[[], Any] = lambda: lowlevel.checkpoint()


@export_module("autogen.agentchat.realtime.experimental")
class RealtimeAgent:
    def __init__(
        self,
        *,
        name: str,
        audio_adapter: Optional[RealtimeObserver] = None,
        system_message: str = "You are a helpful AI Assistant.",
        llm_config: Optional[Union[LLMConfig, dict[str, Any]]] = None,
        logger: Optional[Logger] = None,
        observers: Optional[list[RealtimeObserver]] = None,
        **client_kwargs: Any,
    ):
        """(Experimental) Agent for interacting with the Realtime Clients.

        Args:
            name (str): The name of the agent.
            audio_adapter (Optional[RealtimeObserver] = None): The audio adapter for the agent.
            system_message (str): The system message for the agent.
            llm_config (LLMConfig, dict[str, Any], bool): The config for the agent.
            logger (Optional[Logger]): The logger for the agent.
            observers (Optional[list[RealtimeObserver]]): The additional observers for the agent.
            **client_kwargs (Any): The keyword arguments for the client.
        """
        self._logger = logger
        self._name = name
        self._system_message = system_message

        llm_config = LLMConfig.get_current_llm_config(llm_config)

        self._realtime_client: RealtimeClientProtocol = get_client(
            llm_config=llm_config, logger=self.logger, **client_kwargs
        )

        self._registered_realtime_tools: dict[str, Tool] = {}
        self._observers: list[RealtimeObserver] = observers if observers else []
        self._observers.append(FunctionObserver(logger=logger))
        if audio_adapter:
            self._observers.append(audio_adapter)

        self.callbacks = RealtimeAgentCallbacks()

    @property
    def system_message(self) -> str:
        """Get the system message for the agent."""
        return self._system_message

    @property
    def logger(self) -> Logger:
        """Get the logger for the agent."""
        return self._logger or global_logger

    @property
    def realtime_client(self) -> RealtimeClientProtocol:
        """Get the OpenAI Realtime Client."""
        return self._realtime_client

    @property
    def registered_realtime_tools(self) -> dict[str, Tool]:
        """Get the registered realtime tools."""
        return self._registered_realtime_tools

    def register_observer(self, observer: RealtimeObserver) -> None:
        """Register an observer with the Realtime Agent.

        Args:
            observer (RealtimeObserver): The observer to register.
        """
        self._observers.append(observer)

    async def start_observers(self) -> None:
        for observer in self._observers:
            self._tg.soonify(observer.run)(self)

        # wait for the observers to be ready
        for observer in self._observers:
            await observer.wait_for_ready()

        await self.callbacks.on_observers_ready()

    async def run(self) -> None:
        """Run the agent."""
        # everything is run in the same task group to enable easy cancellation using self._tg.cancel_scope.cancel()
        async with create_task_group() as self._tg:  # noqa: SIM117
            # connect with the client first (establishes a connection and initializes a session)
            async with self._realtime_client.connect():
                # start the observers and wait for them to be ready
                await self.realtime_client.session_update(session_options={"instructions": self.system_message})
                await self.start_observers()

                # iterate over the events
                async for event in self.realtime_client.read_events():
                    for observer in self._observers:
                        await observer.on_event(event)

    def register_realtime_function(
        self,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Callable[[Union[F, Tool]], Tool]:
        """Decorator for registering a function to be used by an agent.

        Args:
            name (str): The name of the function.
            description (str): The description of the function.

        Returns:
            Callable[[Union[F, Tool]], Tool]: The decorator for registering a function.
        """

        def _decorator(func_or_tool: Union[F, Tool]) -> Tool:
            """Decorator for registering a function to be used by an agent.

            Args:
                func_or_tool (Union[F, Tool]): The function or tool to register.

            Returns:
                Tool: The registered tool.
            """
            tool = Tool(func_or_tool=func_or_tool, name=name, description=description)

            self._registered_realtime_tools[tool.name] = tool

            return tool

        return _decorator



# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0

import json
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from logging import Logger, getLogger
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

from ......doc_utils import export_module
from ......import_utils import optional_import_block, require_optional_import
from ......llm_config import LLMConfig
from ...realtime_events import AudioDelta, FunctionCall, RealtimeEvent, SessionCreated
from ..realtime_client import RealtimeClientBase, Role, register_realtime_client

with optional_import_block():
    from websockets.asyncio.client import connect


if TYPE_CHECKING:
    from websockets.asyncio.client import ClientConnection

    from ..realtime_client import RealtimeClientProtocol

__all__ = ["GeminiRealtimeClient"]

global_logger = getLogger(__name__)


HOST = "generativelanguage.googleapis.com"
API_VERSION = "v1alpha"


@register_realtime_client()
@require_optional_import("websockets", "gemini", except_for=["get_factory", "__init__"])
@export_module("autogen.agentchat.realtime.experimental.clients")
class GeminiRealtimeClient(RealtimeClientBase):
    """(Experimental) Client for Gemini Realtime API."""

    def __init__(
        self,
        *,
        llm_config: Union[LLMConfig, dict[str, Any]],
        logger: Optional[Logger] = None,
    ) -> None:
        """(Experimental) Client for Gemini Realtime API.

        Args:
            llm_config: The config for the client.
            logger: The logger for the client.
        """
        super().__init__()
        self._llm_config = llm_config
        self._logger = logger

        self._connection: Optional["ClientConnection"] = None
        config = llm_config["config_list"][0]

        self._model: str = config["model"]
        self._voice = config.get("voice", "charon")
        self._temperature: float = config.get("temperature", 0.8)  # type: ignore[union-attr]

        self._response_modality = "AUDIO"

        self._api_key = config.get("api_key", None)
        # todo: add test with base_url just to make sure it works
        self._base_url: str = config.get(
            "base_url",
            f"wss://{HOST}/ws/google.ai.generativelanguage.{API_VERSION}.GenerativeService.BidiGenerateContent?key={self._api_key}",
        )
        self._final_config: dict[str, Any] = {}
        self._pending_session_updates: dict[str, Any] = {}
        self._is_reading_events = False

    @property
    def logger(self) -> Logger:
        """Get the logger for the Gemini Realtime API."""
        return self._logger or global_logger

    @property
    def connection(self) -> "ClientConnection":
        """Get the Gemini WebSocket connection."""
        if self._connection is None:
            raise RuntimeError("Gemini WebSocket is not initialized")
        return self._connection

    async def send_function_result(self, call_id: str, result: str) -> None:
        """Send the result of a function call to the Gemini Realtime API.

        Args:
            call_id (str): The ID of the function call.
            result (str): The result of the function call.
        """
        msg = {
            "tool_response": {"function_responses": [{"id": call_id, "response": {"result": {"string_value": result}}}]}
        }
        if self._is_reading_events:
            await self.connection.send(json.dumps(msg))

    async def send_text(self, *, role: Role, text: str, turn_complete: bool = True) -> None:
        """Send a text message to the Gemini Realtime API.

        Args:
            role: The role of the message.
            text: The text of the message.
            turn_complete: A flag indicating if the turn is complete.
        """
        msg = {
            "client_content": {
                "turn_complete": turn_complete,
                "turns": [{"role": role, "parts": [{"text": text}]}],
            }
        }
        if self._is_reading_events:
            await self.connection.send(json.dumps(msg))

    async def send_audio(self, audio: str) -> None:
        """Send audio to the Gemini Realtime API.

        Args:
            audio (str): The audio to send.
        """
        msg = {
            "realtime_input": {
                "media_chunks": [
                    {
                        "data": audio,
                        "mime_type": "audio/pcm",
                    }
                ]
            }
        }
        await self.queue_input_audio_buffer_delta(audio)
        if self._is_reading_events:
            await self.connection.send(json.dumps(msg))

    async def truncate_audio(self, audio_end_ms: int, content_index: int, item_id: str) -> None:
        self.logger.info("This is not natively supported by Gemini Realtime API.")
        pass

    async def _initialize_session(self) -> None:
        """Initialize the session with the Gemini Realtime API."""
        session_config = {
            "setup": {
                "system_instruction": {
                    "role": "system",
                    "parts": [{"text": self._pending_session_updates.get("instructions", "")}],
                },
                "model": f"models/{self._model}",
                "tools": [
                    {
                        "function_declarations": [
                            {
                                "name": tool_schema["name"],
                                "description": tool_schema["description"],
                                "parameters": tool_schema["parameters"],
                            }
                            for tool_schema in self._pending_session_updates.get("tools", [])
                        ]
                    },
                ],
                "generation_config": {
                    "response_modalities": [self._response_modality],
                    "speech_config": {"voiceConfig": {"prebuiltVoiceConfig": {"voiceName": self._voice}}},
                    "temperature": self._temperature,
                },
            }
        }

        self.logger.info(f"Sending session update: {session_config}")
        await self.connection.send(json.dumps(session_config))

    async def session_update(self, session_options: dict[str, Any]) -> None:
        """Record session updates to be applied when the connection is established.

        Args:
            session_options (dict[str, Any]): The session options to update.
        """
        if self._is_reading_events:
            self.logger.warning("Is reading events. Session update will be ignored.")
        else:
            self._pending_session_updates.update(session_options)

    @asynccontextmanager
    async def connect(self) -> AsyncGenerator[None, None]:
        """Connect to the Gemini Realtime API."""
        try:
            async with connect(
                self._base_url, additional_headers={"Content-Type": "application/json"}
            ) as self._connection:
                yield
        finally:
            self._connection = None

    async def read_events(self) -> AsyncGenerator[RealtimeEvent, None]:
        """Read Events from the Gemini Realtime Client"""
        if self._connection is None:
            raise RuntimeError("Client is not connected, call connect() first.")
        await self._initialize_session()

        self._is_reading_events = True

        async for event in self._read_events():
            yield event

    async def _read_from_connection(self) -> AsyncGenerator[RealtimeEvent, None]:
        """Read messages from the Gemini Realtime connection."""
        async for raw_message in self.connection:
            message = raw_message.decode("ascii") if isinstance(raw_message, bytes) else raw_message
            events = self._parse_message(json.loads(message))
            for event in events:
                yield event

    def _parse_message(self, response: dict[str, Any]) -> list[RealtimeEvent]:
        """Parse a message from the Gemini Realtime API.

        Args:
            response (dict[str, Any]): The response to parse.

        Returns:
            list[RealtimeEvent]: The parsed events.
        """
        if "serverContent" in response and "modelTurn" in response["serverContent"]:
            try:
                b64data = response["serverContent"]["modelTurn"]["parts"][0]["inlineData"].pop("data")
                return [
                    AudioDelta(
                        delta=b64data,
                        item_id=None,
                        raw_message=response,
                    )
                ]
            except KeyError:
                return []
        elif "toolCall" in response:
            return [
                FunctionCall(
                    raw_message=response,
                    call_id=call["id"],
                    name=call["name"],
                    arguments=call["args"],
                )
                for call in response["toolCall"]["functionCalls"]
            ]
        elif "setupComplete" in response:
            return [
                SessionCreated(raw_message=response),
            ]
        else:
            return [RealtimeEvent(raw_message=response)]

    @classmethod
    def get_factory(
        cls, llm_config: Union[LLMConfig, dict[str, Any]], logger: Logger, **kwargs: Any
    ) -> Optional[Callable[[], "RealtimeClientProtocol"]]:
        """Create a Realtime API client.

        Args:
            llm_config: The LLM config for the client.
            logger: The logger for the client.
            **kwargs: Additional arguments.

        Returns:
            RealtimeClientProtocol: The Realtime API client is returned if the model matches the pattern
        """
        if llm_config["config_list"][0].get("api_type") == "google" and list(kwargs.keys()) == []:
            return lambda: GeminiRealtimeClient(llm_config=llm_config, logger=logger, **kwargs)
        return None


# needed for mypy to check if GeminiRealtimeClient implements RealtimeClientProtocol
if TYPE_CHECKING:
    _client: RealtimeClientProtocol = GeminiRealtimeClient(llm_config={})



import asyncio
import base64
import os
import shlex
import shutil
from enum import Enum
from pathlib import Path
from typing import Literal, TypedDict, Optional, Tuple
from uuid import uuid4

from anthropic.types.beta import BetaToolComputerUse20241022Param

from .base import BaseAnthropicTool, ToolError, ToolResult
from .run import run

OUTPUT_DIR = "/tmp/outputs"

TYPING_DELAY_MS = 12
TYPING_GROUP_SIZE = 50

Action = Literal[
    "key",
    "type",
    "mouse_move",
    "left_click",
    "left_click_drag",
    "right_click",
    "middle_click",
    "double_click",
    "screenshot",
    "cursor_position",
]


class Resolution(TypedDict):
    width: int
    height: int


# sizes above XGA/WXGA are not recommended (see README.md)
# scale down to one of these targets if ComputerTool._scaling_enabled is set
MAX_SCALING_TARGETS: dict[str, Resolution] = {
    "XGA": Resolution(width=1024, height=768),  # 4:3
    "WXGA": Resolution(width=1280, height=800),  # 16:10
    "FWXGA": Resolution(width=1366, height=768),  # ~16:9
}


class ScalingSource(Enum):
    COMPUTER = "computer"
    API = "api"


class ComputerToolOptions(TypedDict):
    display_height_px: int
    display_width_px: int
    display_number: Optional[int]


def chunks(s: str, chunk_size: int) -> list[str]:
    return [s[i : i + chunk_size] for i in range(0, len(s), chunk_size)]


class ComputerTool(BaseAnthropicTool):
    """
    A tool that allows the agent to interact with the screen, keyboard, and mouse of the current computer.
    The tool parameters are defined by Anthropic and are not editable.
    """

    name: Literal["computer"] = "computer"
    api_type: Literal["computer_20241022"] = "computer_20241022"
    width: int
    height: int
    display_num: Optional[int]

    _screenshot_delay = 2.0
    _scaling_enabled = True

    @property
    def options(self) -> ComputerToolOptions:
        width, height = self.scale_coordinates(
            ScalingSource.COMPUTER, self.width, self.height
        )
        return {
            "display_width_px": width,
            "display_height_px": height,
            "display_number": self.display_num,
        }

    def to_params(self) -> BetaToolComputerUse20241022Param:
        return {"name": self.name, "type": self.api_type, **self.options}

    def __init__(self):
        super().__init__()

        self.width = int(os.getenv("WIDTH") or 0)
        self.height = int(os.getenv("HEIGHT") or 0)
        assert self.width and self.height, "WIDTH, HEIGHT must be set"
        if (display_num := os.getenv("DISPLAY_NUM")) is not None:
            self.display_num = int(display_num)
            self._display_prefix = f"DISPLAY=:{self.display_num} "
        else:
            self.display_num = None
            self._display_prefix = ""

        self.xdotool = f"{self._display_prefix}xdotool"

    async def __call__(
        self,
        *,
        action: Action,
        text: Optional[str] = None,
        coordinate: Optional[Tuple[int, int]] = None,
        **kwargs,
    ):
        if action in ("mouse_move", "left_click_drag"):
            if coordinate is None:
                raise ToolError(f"coordinate is required for {action}")
            if text is not None:
                raise ToolError(f"text is not accepted for {action}")
            if not isinstance(coordinate, list) or len(coordinate) != 2:
                raise ToolError(f"{coordinate} must be a tuple of length 2")
            if not all(isinstance(i, int) and i >= 0 for i in coordinate):
                raise ToolError(f"{coordinate} must be a tuple of non-negative ints")

            x, y = self.scale_coordinates(
                ScalingSource.API, coordinate[0], coordinate[1]
            )

            if action == "mouse_move":
                return await self.shell(f"{self.xdotool} mousemove --sync {x} {y}")
            elif action == "left_click_drag":
                return await self.shell(
                    f"{self.xdotool} mousedown 1 mousemove --sync {x} {y} mouseup 1"
                )

        if action in ("key", "type"):
            if text is None:
                raise ToolError(f"text is required for {action}")
            if coordinate is not None:
                raise ToolError(f"coordinate is not accepted for {action}")
            if not isinstance(text, str):
                raise ToolError(output=f"{text} must be a string")

            if action == "key":
                return await self.shell(f"{self.xdotool} key -- {text}")
            elif action == "type":
                results: list[ToolResult] = []
                for chunk in chunks(text, TYPING_GROUP_SIZE):
                    cmd = f"{self.xdotool} type --delay {TYPING_DELAY_MS} -- {shlex.quote(chunk)}"
                    results.append(await self.shell(cmd, take_screenshot=False))
                screenshot_base64 = (await self.screenshot()).base64_image
                return ToolResult(
                    output="".join(result.output or "" for result in results),
                    error="".join(result.error or "" for result in results),
                    base64_image=screenshot_base64,
                )

        if action in (
            "left_click",
            "right_click",
            "double_click",
            "middle_click",
            "screenshot",
            "cursor_position",
        ):
            if text is not None:
                raise ToolError(f"text is not accepted for {action}")
            if coordinate is not None:
                raise ToolError(f"coordinate is not accepted for {action}")

            if action == "screenshot":
                return await self.screenshot()
            elif action == "cursor_position":
                result = await self.shell(
                    f"{self.xdotool} getmouselocation --shell",
                    take_screenshot=False,
                )
                output = result.output or ""
                x, y = self.scale_coordinates(
                    ScalingSource.COMPUTER,
                    int(output.split("X=")[1].split("\n")[0]),
                    int(output.split("Y=")[1].split("\n")[0]),
                )
                return result.replace(output=f"X={x},Y={y}")
            else:
                click_arg = {
                    "left_click": "1",
                    "right_click": "3",
                    "middle_click": "2",
                    "double_click": "--repeat 2 --delay 500 1",
                }[action]
                return await self.shell(f"{self.xdotool} click {click_arg}")

        raise ToolError(f"Invalid action: {action}")

    async def screenshot(self):
        """Take a screenshot of the current screen and return the base64 encoded image."""
        output_dir = Path(OUTPUT_DIR)
        output_dir.mkdir(parents=True, exist_ok=True)
        path = output_dir / f"screenshot_{uuid4().hex}.png"

        # Try gnome-screenshot first
        if shutil.which("gnome-screenshot"):
            screenshot_cmd = f"{self._display_prefix}gnome-screenshot -f {path} -p"
        else:
            # Fall back to scrot if gnome-screenshot isn't available
            screenshot_cmd = f"{self._display_prefix}scrot -p {path}"

        result = await self.shell(screenshot_cmd, take_screenshot=False)
        if self._scaling_enabled:
            x, y = self.scale_coordinates(
                ScalingSource.COMPUTER, self.width, self.height
            )
            await self.shell(
                f"convert {path} -resize {x}x{y}! {path}", take_screenshot=False
            )

        if path.exists():
            return result.replace(
                base64_image=base64.b64encode(path.read_bytes()).decode()
            )
        raise ToolError(f"Failed to take screenshot: {result.error}")

    async def shell(self, command: str, take_screenshot=True) -> ToolResult:
        """Run a shell command and return the output, error, and optionally a screenshot."""
        _, stdout, stderr = await run(command)
        base64_image = None

        if take_screenshot:
            # delay to let things settle before taking a screenshot
            await asyncio.sleep(self._screenshot_delay)
            base64_image = (await self.screenshot()).base64_image

        return ToolResult(output=stdout, error=stderr, base64_image=base64_image)

    def scale_coordinates(self, source: ScalingSource, x: int, y: int):
        """Scale coordinates to a target maximum resolution."""
        if not self._scaling_enabled:
            return x, y
        ratio = self.width / self.height
        target_dimension = None
        for dimension in MAX_SCALING_TARGETS.values():
            # allow some error in the aspect ratio - not ratios are exactly 16:9
            if abs(dimension["width"] / dimension["height"] - ratio) < 0.02:
                if dimension["width"] < self.width:
                    target_dimension = dimension
                break
        if target_dimension is None:
            return x, y
        # should be less than 1
        x_scaling_factor = target_dimension["width"] / self.width
        y_scaling_factor = target_dimension["height"] / self.height
        if source == ScalingSource.API:
            if x > self.width or y > self.height:
                raise ToolError(f"Coordinates {x}, {y} are out of bounds")
            # scale up
            return round(x / x_scaling_factor), round(y / y_scaling_factor)
        # scale down
        return round(x * x_scaling_factor), round(y * y_scaling_factor)


# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0

from collections.abc import AsyncIterator
from typing import Any, Protocol, runtime_checkable

__all__ = ["WebSocketProtocol"]


@runtime_checkable
class WebSocketProtocol(Protocol):
    """WebSocket protocol for sending and receiving JSON data modelled after FastAPI's WebSocket."""

    async def send_json(self, data: Any, mode: str = "text") -> None: ...

    async def receive_json(self, mode: str = "text") -> Any: ...

    async def receive_text(self) -> str: ...

    def iter_text(self) -> AsyncIterator[str]: ...



# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
from datetime import datetime, timezone
from typing import Annotated, Any, Union

from .....doc_utils import export_module
from .....import_utils import optional_import_block, require_optional_import
from .... import Tool
from ....dependency_injection import Depends, on

__all__ = ["DiscordRetrieveTool", "DiscordSendTool"]

with optional_import_block():
    from discord import Client, Intents, utils

MAX_MESSAGE_LENGTH = 2000
MAX_BATCH_RETRIEVE_MESSAGES = 100  # Discord's max per request


@require_optional_import(["discord"], "commsagent-discord")
@export_module("autogen.tools.experimental")
class DiscordSendTool(Tool):
    """Sends a message to a Discord channel."""

    def __init__(self, *, bot_token: str, channel_name: str, guild_name: str) -> None:
        """
        Initialize the DiscordSendTool.

        Args:
            bot_token: The bot token to use for sending messages.
            channel_name: The name of the channel to send messages to.
            guild_name: The name of the guild for the channel.
        """

        # Function that sends the message, uses dependency injection for bot token / channel / guild
        async def discord_send_message(
            message: Annotated[str, "Message to send to the channel."],
            bot_token: Annotated[str, Depends(on(bot_token))],
            guild_name: Annotated[str, Depends(on(guild_name))],
            channel_name: Annotated[str, Depends(on(channel_name))],
        ) -> Any:
            """
            Sends a message to a Discord channel.

            Args:
                message: The message to send to the channel.
                bot_token: The bot token to use for Discord. (uses dependency injection)
                guild_name: The name of the server. (uses dependency injection)
                channel_name: The name of the channel. (uses dependency injection)
            """
            intents = Intents.default()
            intents.message_content = True
            intents.guilds = True
            intents.guild_messages = True

            client = Client(intents=intents)
            result_future: asyncio.Future[str] = asyncio.Future()  # Stores the result of the send

            # When the client is ready, we'll send the message
            @client.event  # type: ignore[misc]
            async def on_ready() -> None:
                try:
                    # Server
                    guild = utils.get(client.guilds, name=guild_name)
                    if guild:
                        # Channel
                        channel = utils.get(guild.text_channels, name=channel_name)
                        if channel:
                            # Send the message
                            if len(message) > MAX_MESSAGE_LENGTH:
                                chunks = [
                                    message[i : i + (MAX_MESSAGE_LENGTH - 1)]
                                    for i in range(0, len(message), (MAX_MESSAGE_LENGTH - 1))
                                ]
                                for i, chunk in enumerate(chunks):
                                    sent = await channel.send(chunk)

                                    # Store ID for the first chunk
                                    if i == 0:
                                        sent_message_id = str(sent.id)

                                result_future.set_result(
                                    f"Message sent successfully ({len(chunks)} chunks, first ID: {sent_message_id}):\n{message}"
                                )
                            else:
                                sent = await channel.send(message)
                                result_future.set_result(f"Message sent successfully (ID: {sent.id}):\n{message}")
                        else:
                            result_future.set_result(f"Message send failed, could not find channel: {channel_name}")
                    else:
                        result_future.set_result(f"Message send failed, could not find guild: {guild_name}")

                except Exception as e:
                    result_future.set_exception(e)
                finally:
                    try:
                        await client.close()
                    except Exception as e:
                        raise Exception(f"Unable to close Discord client: {e}")

            # Start the client and when it's ready it'll send the message in on_ready
            try:
                await client.start(bot_token)

                # Capture the result of the send
                return await result_future
            except Exception as e:
                raise Exception(f"Failed to start Discord client: {e}")

        super().__init__(
            name="discord_send",
            description="Sends a message to a Discord channel.",
            func_or_tool=discord_send_message,
        )


@require_optional_import(["discord"], "commsagent-discord")
@export_module("autogen.tools.experimental")
class DiscordRetrieveTool(Tool):
    """Retrieves messages from a Discord channel."""

    def __init__(self, *, bot_token: str, channel_name: str, guild_name: str) -> None:
        """
        Initialize the DiscordRetrieveTool.

        Args:
            bot_token: The bot token to use for retrieving messages.
            channel_name: The name of the channel to retrieve messages from.
            guild_name: The name of the guild for the channel.
        """

        async def discord_retrieve_messages(
            bot_token: Annotated[str, Depends(on(bot_token))],
            guild_name: Annotated[str, Depends(on(guild_name))],
            channel_name: Annotated[str, Depends(on(channel_name))],
            messages_since: Annotated[
                Union[str, None],
                "Date to retrieve messages from (ISO format) OR Discord snowflake ID. If None, retrieves latest messages.",
            ] = None,
            maximum_messages: Annotated[
                Union[int, None], "Maximum number of messages to retrieve. If None, retrieves all messages since date."
            ] = None,
        ) -> Any:
            """
            Retrieves messages from a Discord channel.

            Args:
                bot_token: The bot token to use for Discord. (uses dependency injection)
                guild_name: The name of the server. (uses dependency injection)
                channel_name: The name of the channel. (uses dependency injection)
                messages_since: ISO format date string OR Discord snowflake ID, to retrieve messages from. If None, retrieves latest messages.
                maximum_messages: Maximum number of messages to retrieve. If None, retrieves all messages since date.
            """
            intents = Intents.default()
            intents.message_content = True
            intents.guilds = True
            intents.guild_messages = True

            client = Client(intents=intents)
            result_future: asyncio.Future[list[dict[str, Any]]] = asyncio.Future()

            messages_since_date: Union[str, None] = None
            if messages_since is not None:
                if DiscordRetrieveTool._is_snowflake(messages_since):
                    messages_since_date = DiscordRetrieveTool._snowflake_to_iso(messages_since)
                else:
                    messages_since_date = messages_since

            @client.event  # type: ignore[misc]
            async def on_ready() -> None:
                try:
                    messages = []

                    # Get guild and channel
                    guild = utils.get(client.guilds, name=guild_name)
                    if not guild:
                        result_future.set_result([{"error": f"Could not find guild: {guild_name}"}])
                        return

                    channel = utils.get(guild.text_channels, name=channel_name)
                    if not channel:
                        result_future.set_result([{"error": f"Could not find channel: {channel_name}"}])
                        return

                    # Setup retrieval parameters
                    last_message_id = None
                    messages_retrieved = 0

                    # Convert to ISO format
                    after_date = None
                    if messages_since_date:
                        try:
                            from datetime import datetime

                            after_date = datetime.fromisoformat(messages_since_date)
                        except ValueError:
                            result_future.set_result([
                                {"error": f"Invalid date format: {messages_since_date}. Use ISO format."}
                            ])
                            return

                    while True:
                        # Setup fetch options
                        fetch_options = {
                            "limit": MAX_BATCH_RETRIEVE_MESSAGES,
                            "before": last_message_id if last_message_id else None,
                            "after": after_date if after_date else None,
                        }

                        # Fetch batch of messages
                        message_batch = []
                        async for message in channel.history(**fetch_options):  # type: ignore[arg-type]
                            message_batch.append(message)
                            messages_retrieved += 1

                            # Check if we've reached the maximum
                            if maximum_messages and messages_retrieved >= maximum_messages:
                                break

                        if not message_batch:
                            break

                        # Process messages
                        for msg in message_batch:
                            messages.append({
                                "id": str(msg.id),
                                "content": msg.content,
                                "author": str(msg.author),
                                "timestamp": msg.created_at.isoformat(),
                            })

                        # Update last message ID for pagination
                        last_message_id = message_batch[-1]  # Use message object directly as 'before' parameter

                        # Break if we've reached the maximum
                        if maximum_messages and messages_retrieved >= maximum_messages:
                            break

                    result_future.set_result(messages)

                except Exception as e:
                    result_future.set_exception(e)
                finally:
                    try:
                        await client.close()
                    except Exception as e:
                        raise Exception(f"Unable to close Discord client: {e}")

            try:
                await client.start(bot_token)
                return await result_future
            except Exception as e:
                raise Exception(f"Failed to start Discord client: {e}")

        super().__init__(
            name="discord_retrieve",
            description="Retrieves messages from a Discord channel based datetime/message ID and/or number of latest messages.",
            func_or_tool=discord_retrieve_messages,
        )

    @staticmethod
    def _is_snowflake(value: str) -> bool:
        """Check if a string is a valid Discord snowflake ID."""
        # Must be numeric and 17-20 digits
        if not value.isdigit():
            return False

        digit_count = len(value)
        return 17 <= digit_count <= 20

    @staticmethod
    def _snowflake_to_iso(snowflake: str) -> str:
        """Convert a Discord snowflake ID to ISO timestamp string."""
        if not DiscordRetrieveTool._is_snowflake(snowflake):
            raise ValueError(f"Invalid snowflake ID: {snowflake}")

        # Discord epoch (2015-01-01)
        discord_epoch = 1420070400000

        # Convert ID to int and shift right 22 bits to get timestamp
        timestamp_ms = (int(snowflake) >> 22) + discord_epoch

        # Convert to datetime and format as ISO string
        dt = datetime.fromtimestamp(timestamp_ms / 1000.0, tz=timezone.utc)
        return dt.isoformat()



import os
import threading
from typing import Dict, List, Set


class PromptRegistry:
    """Central registry for system and module prompts.

    - Loads .txt prompt files from the local `system` and `module` directories on initialization
    - Allows programmatic registration of prompts (module-level or dynamic)
    - Provides a simple, thread-safe API for getting/setting/listing prompts
    - Names for module prompts are based on relative path inside `module`, e.g. `manager/task_planner`
    """

    def __init__(self):
        self._lock = threading.Lock()
        self._prompts: Dict[str, str] = {}
        self._base_dir = os.path.dirname(__file__)
        self._system_dir = os.path.join(self._base_dir, "system")
        self._module_dir = os.path.join(self._base_dir, "module")
        # Track namespaces
        self._module_keys: Set[str] = set()
        self._system_keys: Set[str] = set()
        self._load_system_prompts()
        self._load_module_prompts()

    def _load_system_prompts(self) -> None:
        system_prompts: Dict[str, str] = {}
        system_keys: Set[str] = set()
        try:
            if os.path.isdir(self._system_dir):
                for fname in os.listdir(self._system_dir):
                    if not fname.lower().endswith(".txt"):
                        continue
                    key = os.path.splitext(fname)[0]
                    fpath = os.path.join(self._system_dir, fname)
                    try:
                        with open(fpath, "r", encoding="utf-8") as f:
                            system_prompts[key] = f.read()
                            system_keys.add(key)
                    except Exception:
                        # Skip unreadable files but continue loading others
                        continue
        finally:
            with self._lock:
                # System prompts become the baseline; module/dynamic can override
                self._prompts.update(system_prompts)
                self._system_keys = system_keys

    def _load_module_prompts(self) -> None:
        module_prompts: Dict[str, str] = {}
        module_keys: Set[str] = set()
        try:
            if os.path.isdir(self._module_dir):
                for root, _dirs, files in os.walk(self._module_dir):
                    for fname in files:
                        if not fname.lower().endswith(".txt"):
                            continue
                        fpath = os.path.join(root, fname)
                        rel = os.path.relpath(fpath, self._module_dir)
                        key = os.path.splitext(rel)[0].replace(os.sep, "/")
                        try:
                            with open(fpath, "r", encoding="utf-8") as f:
                                module_prompts[key] = f.read()
                                module_keys.add(key)
                        except Exception:
                            continue
        finally:
            with self._lock:
                # Module prompts can override system prompts of the same key
                self._prompts.update(module_prompts)
                self._module_keys = module_keys

    def refresh(self) -> None:
        """Reload system and module prompts from disk. Keeps any programmatically registered prompts unless overwritten by disk files."""
        with self._lock:
            old_prompts = dict(self._prompts)
        # Reload from disk
        with self._lock:
            self._prompts = {}
            self._module_keys = set()
            self._system_keys = set()
        self._load_system_prompts()
        self._load_module_prompts()
        # Reapply dynamic prompts that are not present on disk
        with self._lock:
            for name, content in old_prompts.items():
                if name not in self._prompts:
                    self._prompts[name] = content

    def _names_from_dir(self, directory: str) -> List[str]:
        if not os.path.isdir(directory):
            return []
        return [os.path.splitext(f)[0] for f in os.listdir(directory) if f.lower().endswith(".txt")]

    def get(self, name: str, default: str = "") -> str:
        with self._lock:
            return self._prompts.get(name, default)

    def set(self, name: str, content: str) -> None:
        """Register or override a prompt by name."""
        with self._lock:
            self._prompts[name] = content

    def exists(self, name: str) -> bool:
        with self._lock:
            return name in self._prompts

    def exists_in_module(self, name: str) -> bool:
        with self._lock:
            return name in self._module_keys

    def module_children_exist(self, prefix: str) -> bool:
        with self._lock:
            prefix_slash = prefix + "/"
            return any(k.startswith(prefix_slash) for k in self._module_keys)

    def all_names(self) -> List[str]:
        with self._lock:
            return sorted(self._prompts.keys())

    def list_by_prefix(self, prefix: str) -> List[str]:
        with self._lock:
            return sorted([n for n in self._prompts.keys() if n.startswith(prefix)])

    def as_dict(self) -> Dict[str, str]:
        with self._lock:
            return dict(self._prompts)


class PromptNamespace:
    """Hierarchical attribute-style access to module prompts.

    Usage:
        from gui_agents.prompts import module
        text = module.evaluator.final_check_role
        text2 = module.manager.planner_role

    Resolution rules:
        - Resolve only against module prompts (under `module/`), ignoring system prompts
        - If an exact module key exists (e.g., "evaluator/final_check_role"), return its string content
        - Else, if there are module children under that path, return a deeper namespace object
        - Else, raise AttributeError
    """

    def __init__(self, registry: PromptRegistry, parts: List[str] = None): #type: ignore
        self._registry = registry
        self._parts = parts or []

    def __getattr__(self, name: str):
        prefix = "/".join(self._parts + [name]) if self._parts else name
        # Exact leaf in module space
        if self._registry.exists_in_module(prefix):
            return self._registry.get(prefix, "")
        # Nested namespace in module space?
        if self._registry.module_children_exist(prefix):
            return PromptNamespace(self._registry, self._parts + [name])
        raise AttributeError(f"No module prompt or namespace '{prefix}'")


# Singleton registry instance for convenient imports
prompt_registry = PromptRegistry()

# Convenience top-level helpers

def get_prompt(name: str, default: str = "") -> str:
    return prompt_registry.get(name, default)


def register_prompt(name: str, content: str) -> None:
    prompt_registry.set(name, content)


def list_prompts() -> List[str]:
    return prompt_registry.all_names()


def list_prompts_by_prefix(prefix: str) -> List[str]:
    return prompt_registry.list_by_prefix(prefix)


def refresh_prompts() -> None:
    prompt_registry.refresh()


# Hierarchical accessor for module prompts
module = PromptNamespace(prompt_registry, []) 


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test script to verify HTML parsing logic for Steam cart evaluation
"""

import json
import sys
import os

# Add the osworld directory to Python path
sys.path.insert(0, '/home/lxguo/osworld')

from desktop_env.evaluators.metrics.chrome import is_added_to_steam_cart
from desktop_env.evaluators.getters.chrome import get_page_info

def test_html_parsing():
    """
    Test the HTML parsing logic for Steam cart evaluation using real HTML file
    """
    print("=== Testing HTML Parsing Logic with Real HTML ===")
    
    # Load the evaluation configuration
    config_path = '/home/lxguo/osworld/evaluation_examples/examples/chrome/121ba48f-9e17-48ce-9bc6-a4fb17a7ebba.json'
    
    try:
        with open(config_path, 'r') as f:
            config = json.load(f)
        print(f"✓ Loaded configuration from {config_path}")
    except Exception as e:
        print(f"✗ Failed to load configuration: {e}")
        return
    
    # Extract evaluation parameters
    evaluator = config['evaluator']
    expected_items = evaluator['expected']['rules']['items']
    target_url = evaluator['result']['url']
    
    print(f"Target URL: {target_url}")
    print(f"Expected items: {expected_items}")
    
    # Load real HTML content from file
    real_html_path = '/home/lxguo/osworld/real_cart.html'
    try:
        with open(real_html_path, 'r', encoding='utf-8') as f:
            real_html_content = f.read()
        print(f"✓ Loaded real HTML from {real_html_path}")
        print(f"HTML content length: {len(real_html_content)} characters")
    except Exception as e:
        print(f"✗ Failed to load real HTML: {e}")
        return
    
    # Simulate active_tab_info structure with real HTML
    real_active_tab_info = {
        'title': 'Shopping Cart',
        'url': target_url,
        'content': real_html_content
    }
    
    # Test the evaluation function with real HTML
    rule = {'items': expected_items}
    
    print("\n=== Testing is_added_to_steam_cart function with Real HTML ===")
    print(f"Rule: {rule}")
    
    # Check if the expected item is in the real HTML content
    print("\n--- Searching for expected items in real HTML ---")
    for item in expected_items:
        if item in real_html_content:
            print(f"✓ Found '{item}' in real HTML content")
        else:
            print(f"✗ '{item}' NOT found in real HTML content")
            
            # Try case-insensitive search
            if item.lower() in real_html_content.lower():
                print(f"  ℹ Found '{item}' with case-insensitive search")
            else:
                print(f"  ℹ '{item}' not found even with case-insensitive search")
                
                # Try searching for individual words
                words = item.split()
                found_words = []
                for word in words:
                    if word.lower() in real_html_content.lower():
                        found_words.append(word)
                if found_words:
                    print(f"  ℹ Found individual words: {found_words}")
                else:
                    print(f"  ℹ No individual words from '{item}' found")
    
    # Call the actual evaluation function
    try:
        result = is_added_to_steam_cart(real_active_tab_info, rule)
        print(f"\nEvaluation result with real HTML: {result}")
        
        if result == 1.0:
            print("✓ Evaluation PASSED - All items found in real HTML")
        else:
            print("✗ Evaluation FAILED - Some items missing in real HTML")
            
    except Exception as e:
        print(f"✗ Error during evaluation: {e}")
        import traceback
        traceback.print_exc()
    
    # Additional analysis of the real HTML content
    print("\n=== Additional Analysis of Real HTML Content ===")
    
    # Check for common Steam cart indicators
    cart_indicators = [
        'cart', 'Cart', 'CART',
        'shopping', 'Shopping', 'SHOPPING',
        'dota', 'Dota', 'DOTA',
        'soundtrack', 'Soundtrack', 'SOUNDTRACK',
        'official', 'Official', 'OFFICIAL'
    ]
    
    print("\n--- Checking for cart-related keywords ---")
    for indicator in cart_indicators:
        count = real_html_content.count(indicator)
        if count > 0:
            print(f"✓ Found '{indicator}' {count} times")
        else:
            print(f"✗ '{indicator}' not found")
    
    # Look for any mention of "Dota" in various forms
    print("\n--- Searching for Dota-related content ---")
    dota_patterns = ['dota', 'Dota', 'DOTA', 'Dota 2', 'dota 2', 'DOTA 2']
    for pattern in dota_patterns:
        if pattern in real_html_content:
            # Find the context around the match
            index = real_html_content.find(pattern)
            start = max(0, index - 50)
            end = min(len(real_html_content), index + len(pattern) + 50)
            context = real_html_content[start:end].replace('\n', ' ').strip()
            print(f"✓ Found '{pattern}' at position {index}")
            print(f"  Context: ...{context}...")
            break
    else:
        print("✗ No Dota-related content found")
    
    # Check if this is actually an empty cart page
    empty_indicators = ['empty', 'Empty', 'EMPTY', 'no items', 'No items', 'NO ITEMS']
    print("\n--- Checking for empty cart indicators ---")
    for indicator in empty_indicators:
        if indicator in real_html_content:
            print(f"⚠ Found empty cart indicator: '{indicator}'")
    
    # Show some basic HTML structure info
    print(f"\n--- HTML Structure Info ---")
    print(f"Total HTML length: {len(real_html_content)} characters")
    print(f"Number of <div> tags: {real_html_content.count('<div')}")
    print(f"Number of <a> tags: {real_html_content.count('<a')}")
    print(f"Contains JavaScript: {'<script' in real_html_content}")
    
    print("\n=== Analysis Complete ===")
    print("\nKey points to check:")
    print("1. The evaluation function uses simple string matching")
    print("2. It checks if ALL items in the rule are present in the HTML content")
    print("3. The HTML content comes from page.content() which returns the full page HTML")
    print("4. If the expected text is not in the HTML, the evaluation will fail")
    print("\nPossible issues:")
    print("- Dynamic content loaded via JavaScript after page.content() is called")
    print("- Text might be in a different format (e.g., encoded, with extra spaces)")
    print("- Page might not be fully loaded when content is captured")
    print("- Network issues preventing proper page loading")

if __name__ == "__main__":
    test_html_parsing()


# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0
#
# Portions derived from  https://github.com/microsoft/autogen are under the MIT License.
# SPDX-License-Identifier: MIT
import logging

from .agentchat import (
    Agent,
    AssistantAgent,
    ChatResult,
    ConversableAgent,
    GroupChat,
    GroupChatManager,
    UpdateSystemMessage,
    UserProxyAgent,
    gather_usage_summary,
    initiate_chats,
    register_function,
)
from .agentchat.group.context_expression import ContextExpression
from .code_utils import DEFAULT_MODEL, FAST_MODEL
from .exception_utils import (
    AgentNameConflictError,
    InvalidCarryOverTypeError,
    NoEligibleSpeakerError,
    SenderRequiredError,
    UndefinedNextAgentError,
)
from .llm_config import LLMConfig
from .oai import (
    Cache,
    ModelClient,
    OpenAIWrapper,
    config_list_from_dotenv,
    config_list_from_json,
    config_list_from_models,
    config_list_gpt4_gpt35,
    config_list_openai_aoai,
    filter_config,
    get_config_list,
)

# Set the root logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


__all__ = [
    "DEFAULT_MODEL",
    "FAST_MODEL",
    "Agent",
    "AgentNameConflictError",
    "AssistantAgent",
    "Cache",
    "ChatResult",
    "ContextExpression",
    "ConversableAgent",
    "GroupChat",
    "GroupChatManager",
    "InvalidCarryOverTypeError",
    "LLMConfig",
    "ModelClient",
    "NoEligibleSpeakerError",
    "OpenAIWrapper",
    "SenderRequiredError",
    "UndefinedNextAgentError",
    "UpdateSystemMessage",
    "UserProxyAgent",
    "config_list_from_dotenv",
    "config_list_from_json",
    "config_list_from_models",
    "config_list_gpt4_gpt35",
    "config_list_openai_aoai",
    "filter_config",
    "gather_usage_summary",
    "get_config_list",
    "initiate_chats",
    "register_function",
]



import os
import logging
import dotenv
import time
import signal
import requests
from datetime import datetime, timedelta, timezone

from alibabacloud_ecs20140526.client import Client as ECSClient
from alibabacloud_tea_openapi import models as open_api_models
from alibabacloud_ecs20140526 import models as ecs_models
from alibabacloud_tea_util.client import Client as UtilClient
from desktop_env.providers.base import VMManager
from desktop_env.providers.aliyun.config import ENABLE_TTL, DEFAULT_TTL_MINUTES


dotenv.load_dotenv()

for env_name in [
    "ALIYUN_REGION",
    "ALIYUN_VSWITCH_ID",
    "ALIYUN_SECURITY_GROUP_ID",
    "ALIYUN_IMAGE_ID",
    "ALIYUN_ACCESS_KEY_ID",
    "ALIYUN_ACCESS_KEY_SECRET",
    "ALIYUN_INSTANCE_TYPE",
]:
    if not os.getenv(env_name):
        raise EnvironmentError(f"{env_name} must be set in the environment variables.")


logger = logging.getLogger("desktopenv.providers.aliyun.AliyunVMManager")
logger.setLevel(logging.INFO)

ALIYUN_INSTANCE_TYPE = os.getenv("ALIYUN_INSTANCE_TYPE")
ALIYUN_ACCESS_KEY_ID = os.getenv("ALIYUN_ACCESS_KEY_ID")
ALIYUN_ACCESS_KEY_SECRET = os.getenv("ALIYUN_ACCESS_KEY_SECRET")
ALIYUN_REGION = os.getenv("ALIYUN_REGION")
ALIYUN_IMAGE_ID = os.getenv("ALIYUN_IMAGE_ID")
ALIYUN_SECURITY_GROUP_ID = os.getenv("ALIYUN_SECURITY_GROUP_ID")
ALIYUN_VSWITCH_ID = os.getenv("ALIYUN_VSWITCH_ID")
ALIYUN_RESOURCE_GROUP_ID = os.getenv("ALIYUN_RESOURCE_GROUP_ID")

WAIT_DELAY = 20
MAX_ATTEMPTS = 15


def _allocate_vm(screen_size=(1920, 1080)):
    """
    Allocate a new Aliyun ECS instance
    """
    assert screen_size == (1920, 1080), "Only 1920x1080 screen size is supported"

    config = open_api_models.Config(
        access_key_id=ALIYUN_ACCESS_KEY_ID,
        access_key_secret=ALIYUN_ACCESS_KEY_SECRET,
        region_id=ALIYUN_REGION,
    )
    client = ECSClient(config)
    instance_id = None
    original_sigint_handler = signal.getsignal(signal.SIGINT)
    original_sigterm_handler = signal.getsignal(signal.SIGTERM)

    def signal_handler(sig, frame):
        if instance_id:
            signal_name = "SIGINT" if sig == signal.SIGINT else "SIGTERM"
            logger.warning(
                f"Received {signal_name} signal, terminating instance {instance_id}..."
            )
            try:
                delete_request = ecs_models.DeleteInstancesRequest(
                    region_id=ALIYUN_REGION,
                    instance_ids=UtilClient.to_jsonstring([instance_id]),
                    force=True,
                )
                client.delete_instances(delete_request)
                logger.info(
                    f"Successfully terminated instance {instance_id} after {signal_name}."
                )
            except Exception as cleanup_error:
                logger.error(
                    f"Failed to terminate instance {instance_id} after {signal_name}: {str(cleanup_error)}"
                )

        # Restore original signal handlers
        signal.signal(signal.SIGINT, original_sigint_handler)
        signal.signal(signal.SIGTERM, original_sigterm_handler)

        # Raise appropriate exception based on signal type
        if sig == signal.SIGINT:
            raise KeyboardInterrupt
        else:
            # For SIGTERM, exit gracefully
            import sys

            sys.exit(0)

    try:
        # Set up signal handlers for both SIGINT and SIGTERM
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        logger.info(
            f"Creating new ECS instance in region {ALIYUN_REGION} with image {ALIYUN_IMAGE_ID}"
        )

        # TTL configuration
        ttl_enabled = ENABLE_TTL
        ttl_minutes = DEFAULT_TTL_MINUTES
        ttl_seconds = max(0, int(ttl_minutes) * 60)

        # Aliyun constraints: at least 30 minutes in the future, ISO8601 UTC, seconds must be 00
        now_utc = datetime.now(timezone.utc)
        min_eta = now_utc + timedelta(minutes=30)
        raw_eta = now_utc + timedelta(seconds=ttl_seconds)
        effective_eta = raw_eta if raw_eta > min_eta else min_eta
        # round up to the next full minute, zero seconds
        effective_eta = (effective_eta + timedelta(seconds=59)).replace(second=0, microsecond=0)
        auto_release_str = effective_eta.strftime('%Y-%m-%dT%H:%M:%SZ')
        logger.info(
            f"TTL config: enabled={ttl_enabled}, minutes={ttl_minutes}, seconds={ttl_seconds}, ETA(UTC)={auto_release_str}"
        )

        # Create instance request (attempt with auto_release_time first when TTL enabled)
        def _build_request(with_ttl: bool) -> ecs_models.RunInstancesRequest:
            kwargs = dict(
                region_id=ALIYUN_REGION,
                image_id=ALIYUN_IMAGE_ID,
                instance_type=ALIYUN_INSTANCE_TYPE,
                security_group_id=ALIYUN_SECURITY_GROUP_ID,
                v_switch_id=ALIYUN_VSWITCH_ID,
                instance_name=f"OSWorld-Desktop-{int(time.time())}",
                description="OSWorld Desktop Environment Instance",
                internet_max_bandwidth_out=10,
                internet_charge_type="PayByTraffic",
                instance_charge_type="PostPaid",
                system_disk=ecs_models.RunInstancesRequestSystemDisk(
                    size="50",
                    category="cloud_essd",
                ),
                deletion_protection=False,
            )
            
            if ALIYUN_RESOURCE_GROUP_ID:
                kwargs["resource_group_id"] = ALIYUN_RESOURCE_GROUP_ID
            
            if with_ttl and ttl_enabled and ttl_seconds > 0:
                kwargs["auto_release_time"] = auto_release_str
            return ecs_models.RunInstancesRequest(**kwargs)

        try:
            request = _build_request(with_ttl=True)
            response = client.run_instances(request)
        except Exception as create_err:
            # Retry without auto_release_time if creation-time TTL is rejected
            logger.warning(
                f"RunInstances with auto_release_time failed: {create_err}. Retrying without TTL field..."
            )
            request = _build_request(with_ttl=False)
            response = client.run_instances(request)
        instance_ids = response.body.instance_id_sets.instance_id_set

        if not instance_ids:
            raise RuntimeError(
                "Failed to create ECS instance - no instance ID returned"
            )

        instance_id = instance_ids[0]
        logger.info(f"ECS instance {instance_id} created successfully")

        # Wait for the instance to be running
        logger.info(f"Waiting for instance {instance_id} to be running...")
        _wait_for_instance_running(client, instance_id)

        logger.info(f"Instance {instance_id} is now running and ready")

    except KeyboardInterrupt:
        logger.warning("VM allocation interrupted by user (SIGINT).")
        if instance_id:
            logger.info(f"Terminating instance {instance_id} due to interruption.")
            try:
                delete_request = ecs_models.DeleteInstancesRequest(
                    region_id=ALIYUN_REGION,
                    instance_ids=UtilClient.to_jsonstring([instance_id]),
                    force=True,
                )
                client.delete_instances(delete_request)
            except Exception as cleanup_error:
                logger.error(
                    f"Failed to cleanup instance {instance_id}: {str(cleanup_error)}"
                )
        raise
    except Exception as e:
        logger.error(f"Failed to allocate ECS instance: {str(e)}")
        if instance_id:
            logger.info(f"Terminating instance {instance_id} due to an error.")
            try:
                delete_request = ecs_models.DeleteInstancesRequest(
                    region_id=ALIYUN_REGION,
                    instance_ids=UtilClient.to_jsonstring([instance_id]),
                    force=True,
                )
                client.delete_instances(delete_request)
            except Exception as cleanup_error:
                logger.error(
                    f"Failed to cleanup instance {instance_id}: {str(cleanup_error)}"
                )
        raise
    finally:
        # Restore original signal handlers
        signal.signal(signal.SIGINT, original_sigint_handler)
        signal.signal(signal.SIGTERM, original_sigterm_handler)

    return instance_id


def _wait_for_instance_running(
    client: ECSClient, instance_id: str, max_attempts: int = MAX_ATTEMPTS
):
    """Wait for instance to reach Running state"""
    for _ in range(max_attempts):
        try:
            req = ecs_models.DescribeInstancesRequest(
                region_id=ALIYUN_REGION,
                instance_ids=UtilClient.to_jsonstring([instance_id]),
            )
            response = client.describe_instances(req)

            if response.body.instances.instance:
                instance = response.body.instances.instance[0]
                status = instance.status
                logger.info(f"Instance {instance_id} status: {status}")

                if status == "Running":
                    return
                elif status in ["Stopped", "Stopping"]:
                    start_req = ecs_models.StartInstanceRequest(instance_id=instance_id)
                    client.start_instance(start_req)
                    logger.info(f"Started instance {instance_id}")

            time.sleep(WAIT_DELAY)

        except Exception as e:
            logger.warning(f"Error checking instance status: {e}")
            time.sleep(WAIT_DELAY)

    raise TimeoutError(
        f"Instance {instance_id} did not reach Running state within {max_attempts * WAIT_DELAY} seconds"
    )


def _wait_until_server_ready(public_ip: str):
    """Wait until the server is ready"""
    for _ in range(MAX_ATTEMPTS):
        try:
            logger.info(f"Checking server status on {public_ip}...")
            response = requests.get(f"http://{public_ip}:5000/", timeout=2)
            if response.status_code == 404:
                logger.info(f"Server {public_ip} is ready")
                return
        except Exception:
            time.sleep(WAIT_DELAY)

    raise TimeoutError(
        f"Server {public_ip} did not respond within {MAX_ATTEMPTS * WAIT_DELAY} seconds"
    )


class AliyunVMManager(VMManager):
    """
    Aliyun ECS VM Manager for managing virtual machines on Aliyun Cloud.

    Aliyun ECS does not need to maintain a registry of VMs, as it can dynamically allocate and deallocate VMs.
    """

    def __init__(self, **kwargs):
        self.initialize_registry()

    def initialize_registry(self, **kwargs):
        pass

    def add_vm(self, vm_path, lock_needed=True, **kwargs):
        pass

    def _add_vm(self, vm_path):
        pass

    def delete_vm(self, vm_path, lock_needed=True, **kwargs):
        pass

    def _delete_vm(self, vm_path):
        pass

    def occupy_vm(self, vm_path, pid, lock_needed=True, **kwargs):
        pass

    def _occupy_vm(self, vm_path, pid):
        pass

    def check_and_clean(self, lock_needed=True, **kwargs):
        pass

    def _check_and_clean(self):
        pass

    def list_free_vms(self, lock_needed=True, **kwargs):
        pass

    def _list_free_vms(self):
        pass

    def get_vm_path(self, screen_size=(1920, 1080), **kwargs):
        """Get a VM path (instance ID) for use"""
        logger.info(
            f"Allocating new ECS instance in region {ALIYUN_REGION} with screen size {screen_size}"
        )

        try:
            instance_id = _allocate_vm(screen_size)
            logger.info(f"Successfully allocated instance {instance_id}")
            return instance_id

        except Exception as e:
            logger.error(f"Failed to allocate instance: {str(e)}")
            raise



#!/usr/bin/env python3
"""
Test the specific FlightAware discussions page that failed in the evaluation
"""

import requests
import json
from urllib.parse import urlparse

def load_proxy_config():
    """Load proxy configuration from dataimpulse.json"""
    try:
        with open('/home/lxguo/osworld/evaluation_examples/settings/proxy/dataimpulse.json', 'r') as f:
            proxy_configs = json.load(f)
            # dataimpulse.json contains an array, get the first proxy
            if isinstance(proxy_configs, list) and len(proxy_configs) > 0:
                return proxy_configs[0]
            else:
                return proxy_configs
    except Exception as e:
        print(f"Error loading proxy config: {e}")
        return None

def get_proxy_dict(proxy_config):
    """Convert proxy config to requests proxy format"""
    if not proxy_config:
        return None
    
    proxy_url = f"http://{proxy_config['username']}:{proxy_config['password']}@{proxy_config['host']}:{proxy_config['port']}"
    return {
        'http': proxy_url,
        'https': proxy_url
    }

def test_specific_url(url, use_proxy=False, proxy_dict=None):
    """Test the specific URL that failed in evaluation"""
    print(f"\n=== Testing {url} ===\n")
    
    try:
        # Test with GET request to simulate browser behavior
        headers = {
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
            'Accept-Language': 'en-US,en;q=0.5',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive',
            'Upgrade-Insecure-Requests': '1'
        }
        
        response = requests.get(
            url, 
            headers=headers,
            proxies=proxy_dict if use_proxy else None,
            timeout=60,
            allow_redirects=True
        )
        
        print(f"Status Code: {response.status_code}")
        print(f"Final URL: {response.url}")
        print(f"Response Size: {len(response.content)} bytes")
        
        # Check response headers
        print("\nKey Response Headers:")
        for header in ['server', 'cf-ray', 'cf-cache-status', 'content-type']:
            if header in response.headers:
                print(f"  {header}: {response.headers[header]}")
        
        # Check if we got the actual page content
        if response.status_code == 200:
            content = response.text.lower()
            if 'banter' in content or 'thread' in content or 'discourse' in content:
                print("✅ Successfully loaded page content")
            elif 'cloudflare' in content or 'challenge' in content:
                print("⚠️  Got Cloudflare challenge page")
            elif len(content) < 1000:
                print(f"⚠️  Suspiciously small content: {len(content)} chars")
                print(f"Content preview: {content[:200]}...")
            else:
                print("✅ Got substantial content")
        
        return True, response.status_code, len(response.content)
        
    except requests.exceptions.ProxyError as e:
        print(f"❌ Proxy Error: {e}")
        return False, None, 0
    except requests.exceptions.Timeout as e:
        print(f"❌ Timeout Error: {e}")
        return False, None, 0
    except requests.exceptions.RequestException as e:
        print(f"❌ Request Error: {e}")
        return False, None, 0

def main():
    print("Testing Specific FlightAware Discussions Page")
    print("=" * 60)
    
    # Load proxy configuration
    proxy_config = load_proxy_config()
    if proxy_config:
        print(f"✅ Loaded proxy: {proxy_config['host']}:{proxy_config['port']}")
        proxy_dict = get_proxy_dict(proxy_config)
    else:
        print("❌ Failed to load proxy configuration")
        return
    
    # The specific URL that failed in the evaluation
    target_url = "https://discussions.flightaware.com/t/the-banter-thread/4412"
    
    print(f"\nTarget URL: {target_url}")
    print("This is the exact URL that failed in the evaluation log\n")
    
    # Test direct connection
    print("=" * 40)
    print("DIRECT CONNECTION TEST")
    print("=" * 40)
    direct_success, direct_status, direct_size = test_specific_url(target_url, use_proxy=False)
    
    # Test proxy connection
    print("\n" + "=" * 40)
    print("PROXY CONNECTION TEST")
    print("=" * 40)
    proxy_success, proxy_status, proxy_size = test_specific_url(target_url, use_proxy=True, proxy_dict=proxy_dict)
    
    # Summary
    print("\n" + "=" * 60)
    print("SUMMARY")
    print("=" * 60)
    
    print(f"\nDirect Connection:")
    print(f"  Success: {'✅' if direct_success else '❌'}")
    print(f"  Status:  {direct_status}")
    print(f"  Size:    {direct_size} bytes")
    
    print(f"\nProxy Connection:")
    print(f"  Success: {'✅' if proxy_success else '❌'}")
    print(f"  Status:  {proxy_status}")
    print(f"  Size:    {proxy_size} bytes")
    
    # Analysis
    print("\n" + "=" * 60)
    print("ANALYSIS")
    print("=" * 60)
    
    if direct_success and not proxy_success:
        print("\n🔍 Finding: Proxy connection fails while direct works")
        print("   This matches the evaluation log error pattern")
    elif not direct_success and not proxy_success:
        print("\n🔍 Finding: Both connections fail")
        print("   The page might be restricted or have issues")
    elif direct_success and proxy_success:
        print("\n🔍 Finding: Both connections work")
        print("   The issue might be intermittent or evaluation-specific")
    else:
        print("\n🔍 Finding: Unexpected pattern - proxy works but direct fails")
    
    # Additional tests with different approaches
    print("\n" + "=" * 60)
    print("ADDITIONAL TESTS")
    print("=" * 60)
    
    # Test the base discussions URL
    base_url = "https://discussions.flightaware.com"
    print(f"\nTesting base URL: {base_url}")
    
    print("\n--- Direct ---")
    base_direct_success, _, _ = test_specific_url(base_url, use_proxy=False)
    
    print("\n--- Proxy ---")
    base_proxy_success, _, _ = test_specific_url(base_url, use_proxy=True, proxy_dict=proxy_dict)
    
    if base_proxy_success and not proxy_success:
        print("\n💡 Insight: Base discussions URL works with proxy, but specific thread fails")
        print("   This suggests the issue is with specific deep-linked content")

if __name__ == "__main__":
    main()


# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0
#
# Portions derived from  https://github.com/microsoft/autogen are under the MIT License.
# SPDX-License-Identifier: MIT

import queue
from asyncio import Queue as AsyncQueue
from typing import Any, AsyncIterable, Dict, Iterable, Optional, Protocol, Sequence, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from autogen.tools.tool import Tool

from ..agentchat.agent import Agent, LLMMessageType
from ..agentchat.group.context_variables import ContextVariables
from ..events.agent_events import ErrorEvent, InputRequestEvent, RunCompletionEvent
from ..events.base_event import BaseEvent
from .processors import (
    AsyncConsoleEventProcessor,
    AsyncEventProcessorProtocol,
    ConsoleEventProcessor,
    EventProcessorProtocol,
)
from .thread_io_stream import AsyncThreadIOStream, ThreadIOStream

Message = dict[str, Any]


class RunInfoProtocol(Protocol):
    @property
    def uuid(self) -> UUID: ...

    @property
    def above_run(self) -> Optional["RunResponseProtocol"]: ...


class Usage(BaseModel):
    cost: float
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class CostBreakdown(BaseModel):
    total_cost: float
    models: Dict[str, Usage] = Field(default_factory=dict)

    @classmethod
    def from_raw(cls, data: dict[str, Any]) -> "CostBreakdown":
        # Extract total cost
        total_cost = data.get("total_cost", 0.0)

        # Remove total_cost key to extract models
        model_usages = {k: Usage(**v) for k, v in data.items() if k != "total_cost"}

        return cls(total_cost=total_cost, models=model_usages)


class Cost(BaseModel):
    usage_including_cached_inference: CostBreakdown
    usage_excluding_cached_inference: CostBreakdown

    @classmethod
    def from_raw(cls, data: dict[str, Any]) -> "Cost":
        return cls(
            usage_including_cached_inference=CostBreakdown.from_raw(data.get("usage_including_cached_inference", {})),
            usage_excluding_cached_inference=CostBreakdown.from_raw(data.get("usage_excluding_cached_inference", {})),
        )


class RunResponseProtocol(RunInfoProtocol, Protocol):
    @property
    def events(self) -> Iterable[BaseEvent]: ...

    @property
    def messages(self) -> Iterable[Message]: ...

    @property
    def summary(self) -> Optional[str]: ...

    @property
    def context_variables(self) -> Optional[ContextVariables]: ...

    @property
    def last_speaker(self) -> Optional[str]: ...

    @property
    def cost(self) -> Optional[Cost]: ...

    def process(self, processor: Optional[EventProcessorProtocol] = None) -> None: ...

    def set_ui_tools(self, tools: list[Tool]) -> None: ...


class AsyncRunResponseProtocol(RunInfoProtocol, Protocol):
    @property
    def events(self) -> AsyncIterable[BaseEvent]: ...

    @property
    async def messages(self) -> Iterable[Message]: ...

    @property
    async def summary(self) -> Optional[str]: ...

    @property
    async def context_variables(self) -> Optional[ContextVariables]: ...

    @property
    async def last_speaker(self) -> Optional[str]: ...

    @property
    async def cost(self) -> Optional[Cost]: ...

    async def process(self, processor: Optional[AsyncEventProcessorProtocol] = None) -> None: ...

    def set_ui_tools(self, tools: list[Tool]) -> None: ...


class RunResponse:
    def __init__(self, iostream: ThreadIOStream, agents: list[Agent]):
        self.iostream = iostream
        self.agents = agents
        self._summary: Optional[str] = None
        self._messages: Sequence[LLMMessageType] = []
        self._uuid = uuid4()
        self._context_variables: Optional[ContextVariables] = None
        self._last_speaker: Optional[str] = None
        self._cost: Optional[Cost] = None

    def _queue_generator(self, q: queue.Queue) -> Iterable[BaseEvent]:  # type: ignore[type-arg]
        """A generator to yield items from the queue until the termination message is found."""
        while True:
            try:
                # Get an item from the queue
                event = q.get(timeout=0.1)  # Adjust timeout as needed

                if isinstance(event, InputRequestEvent):
                    event.content.respond = lambda response: self.iostream._output_stream.put(response)  # type: ignore[attr-defined]

                yield event

                if isinstance(event, RunCompletionEvent):
                    self._messages = event.content.history  # type: ignore[attr-defined]
                    self._last_speaker = event.content.last_speaker  # type: ignore[attr-defined]
                    self._summary = event.content.summary  # type: ignore[attr-defined]
                    self._context_variables = event.content.context_variables  # type: ignore[attr-defined]
                    self.cost = event.content.cost  # type: ignore[attr-defined]
                    break

                if isinstance(event, ErrorEvent):
                    raise event.content.error  # type: ignore[attr-defined]
            except queue.Empty:
                continue  # Wait for more items in the queue

    @property
    def events(self) -> Iterable[BaseEvent]:
        return self._queue_generator(self.iostream.input_stream)

    @property
    def messages(self) -> Iterable[Message]:
        return self._messages

    @property
    def summary(self) -> Optional[str]:
        return self._summary

    @property
    def above_run(self) -> Optional["RunResponseProtocol"]:
        return None

    @property
    def uuid(self) -> UUID:
        return self._uuid

    @property
    def context_variables(self) -> Optional[ContextVariables]:
        return self._context_variables

    @property
    def last_speaker(self) -> Optional[str]:
        return self._last_speaker

    @property
    def cost(self) -> Optional[Cost]:
        return self._cost

    @cost.setter
    def cost(self, value: Union[Cost, dict[str, Any]]) -> None:
        if isinstance(value, dict):
            self._cost = Cost.from_raw(value)
        else:
            self._cost = value

    def process(self, processor: Optional[EventProcessorProtocol] = None) -> None:
        processor = processor or ConsoleEventProcessor()
        processor.process(self)

    def set_ui_tools(self, tools: list[Tool]) -> None:
        """Set the UI tools for the agents."""
        for agent in self.agents:
            agent.set_ui_tools(tools)


class AsyncRunResponse:
    def __init__(self, iostream: AsyncThreadIOStream, agents: list[Agent]):
        self.iostream = iostream
        self.agents = agents
        self._summary: Optional[str] = None
        self._messages: Sequence[LLMMessageType] = []
        self._uuid = uuid4()
        self._context_variables: Optional[ContextVariables] = None
        self._last_speaker: Optional[str] = None
        self._cost: Optional[Cost] = None

    async def _queue_generator(self, q: AsyncQueue[Any]) -> AsyncIterable[BaseEvent]:  # type: ignore[type-arg]
        """A generator to yield items from the queue until the termination message is found."""
        while True:
            try:
                # Get an item from the queue
                event = await q.get()

                if isinstance(event, InputRequestEvent):

                    async def respond(response: str) -> None:
                        await self.iostream._output_stream.put(response)

                    event.content.respond = respond  # type: ignore[attr-defined]

                yield event

                if isinstance(event, RunCompletionEvent):
                    self._messages = event.content.history  # type: ignore[attr-defined]
                    self._last_speaker = event.content.last_speaker  # type: ignore[attr-defined]
                    self._summary = event.content.summary  # type: ignore[attr-defined]
                    self._context_variables = event.content.context_variables  # type: ignore[attr-defined]
                    self.cost = event.content.cost  # type: ignore[attr-defined]
                    break

                if isinstance(event, ErrorEvent):
                    raise event.content.error  # type: ignore[attr-defined]
            except queue.Empty:
                continue

    @property
    def events(self) -> AsyncIterable[BaseEvent]:
        return self._queue_generator(self.iostream.input_stream)

    @property
    async def messages(self) -> Iterable[Message]:
        return self._messages

    @property
    async def summary(self) -> Optional[str]:
        return self._summary

    @property
    def above_run(self) -> Optional["RunResponseProtocol"]:
        return None

    @property
    def uuid(self) -> UUID:
        return self._uuid

    @property
    async def context_variables(self) -> Optional[ContextVariables]:
        return self._context_variables

    @property
    async def last_speaker(self) -> Optional[str]:
        return self._last_speaker

    @property
    async def cost(self) -> Optional[Cost]:
        return self._cost

    @cost.setter
    def cost(self, value: Union[Cost, dict[str, Any]]) -> None:
        if isinstance(value, dict):
            self._cost = Cost.from_raw(value)
        else:
            self._cost = value

    async def process(self, processor: Optional[AsyncEventProcessorProtocol] = None) -> None:
        processor = processor or AsyncConsoleEventProcessor()
        await processor.process(self)

    def set_ui_tools(self, tools: list[Tool]) -> None:
        """Set the UI tools for the agents."""
        for agent in self.agents:
            agent.set_ui_tools(tools)



"""
Component Debugger - Support debugging core components like Worker, Evaluator, Manager, etc.
"""

from typing import Dict, Any, Optional

from pyautogui import screenshot

from ..controller.state_handlers import NewWorker, Evaluator
from ..controller.main_controller import NewManager


class ComponentDebugger:
    """Component Debugger - Support debugging various core components"""
    
    def __init__(self, snapshot_debugger):
        self.snapshot_debugger: SnapshotDebugger = snapshot_debugger
        self.current_worker = None
        self.current_evaluator = None
        self.current_manager = None
        
    def debug_worker(self, snapshot_id: str) -> Optional[NewWorker]:
        """Debug Worker component"""
        if not self.snapshot_debugger.load_snapshot(snapshot_id):
            return None
            
        try:
            worker_params = self.snapshot_debugger.get_worker_params()
            self.current_worker = NewWorker(**worker_params)
            self.current_worker.process_subtask_and_create_command()
            print(f"Worker component created, ready to debug snapshot: {snapshot_id}")
            return self.current_worker
            
        except Exception as e:
            print(f"Failed to create Worker component: {e}")
            return None
    
    def debug_evaluator(self, snapshot_id: str) -> Optional[Evaluator]:
        """Debug Evaluator component"""
        if not self.snapshot_debugger.load_snapshot(snapshot_id):
            return None
            
        try:
            evaluator_params = self.snapshot_debugger.get_evaluator_params()
            self.current_evaluator = Evaluator(**evaluator_params)
            print(f"Evaluator component created, ready to debug snapshot: {snapshot_id}")
            return self.current_evaluator
            
        except Exception as e:
            print(f"Failed to create Evaluator component: {e}")
            return None
    
    def debug_manager(self, snapshot_id: str) -> Optional[NewManager]:
        """Debug Manager component"""
        if not self.snapshot_debugger.load_snapshot(snapshot_id):
            return None
            
        try:
            screenshot = self.snapshot_debugger.global_state.get_screenshot()
            print(f"screenshot: {screenshot.__sizeof__()}")
            manager_params = self.snapshot_debugger.get_manager_params()
            self.current_manager = NewManager(**manager_params)
            self.current_manager.plan_task("replan")
            print(f"Manager component created, ready to debug snapshot: {snapshot_id}")
            return self.current_manager
            
        except Exception as e:
            print(f"Failed to create Manager component: {e}")
            return None
    
    def step_worker(self) -> bool:
        """Step execute Worker"""
        if not self.current_worker:
            print("Please create Worker component first")
            return False
            
        try:
            print("Executing Worker step debugging...")
            self.current_worker.process_subtask_and_create_command()
            print("Worker step execution completed")
            return True
            
        except Exception as e:
            print(f"Worker execution failed: {e}")
            return False
    
    def get_worker_state(self) -> Dict[str, Any]:
        """Get current Worker state"""
        if not self.current_worker:
            return {}
            
        # Here you can get state information based on the actual structure of Worker
        return {
            "worker_created": True,
            "global_state": self.snapshot_debugger.global_state.get_current_state_summary()
        }
    
    def get_evaluator_state(self) -> Dict[str, Any]:
        """Get current Evaluator state"""
        if not self.current_evaluator:
            return {}
            
        return {
            "evaluator_created": True,
            "global_state": self.snapshot_debugger.global_state.get_current_state_summary()
        }
    
    def get_manager_state(self) -> Dict[str, Any]:
        """Get current Manager state"""
        if not self.current_manager:
            return {}
            
        return {
            "manager_created": True,
            "global_state": self.snapshot_debugger.global_state.get_current_state_summary()
        }
    
    def reset_debug_session(self):
        """Reset debug session"""
        self.current_worker = None
        self.current_evaluator = None
        self.current_manager = None
        print("Debug session has been reset")


import os
import logging
from datetime import datetime

from alibabacloud_ecs20140526.client import Client as ECSClient
from alibabacloud_tea_openapi import models as open_api_models
from alibabacloud_ecs20140526 import models as ecs_models
from alibabacloud_tea_util.client import Client as UtilClient

from desktop_env.providers.base import Provider
from desktop_env.providers.aliyun.manager import (
    _allocate_vm,
    _wait_for_instance_running,
    _wait_until_server_ready,
)


logger = logging.getLogger("desktopenv.providers.aliyun.AliyunProvider")
logger.setLevel(logging.INFO)


class AliyunProvider(Provider):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.region = os.getenv("ALIYUN_REGION", "eu-central-1")
        self.client = self._create_client()
        # Whether to use private IP instead of public IP. Default: enabled.
        # Priority: explicit kwarg > env var ALIYUN_USE_PRIVATE_IP > default True
        env_use_private = os.getenv("ALIYUN_USE_PRIVATE_IP", "1").lower() in {"1", "true", "yes", "on"}
        kw_flag = kwargs.get("use_private_ip", None)
        self.use_private_ip = env_use_private if kw_flag is None else bool(kw_flag)

    def _create_client(self) -> ECSClient:
        config = open_api_models.Config(
            access_key_id=os.getenv("ALIYUN_ACCESS_KEY_ID"),
            access_key_secret=os.getenv("ALIYUN_ACCESS_KEY_SECRET"),
            region_id=self.region,
        )
        return ECSClient(config)

    def start_emulator(self, path_to_vm: str, headless: bool, *args, **kwargs):
        logger.info("Starting Aliyun ECS instance...")

        try:
            # Check the current state of the instance
            response = self._describe_instance(path_to_vm)
            if not response.body.instances.instance:
                logger.error(f"Instance {path_to_vm} not found")
                return

            instance = response.body.instances.instance[0]
            state = instance.status
            logger.info(f"Instance {path_to_vm} current state: {state}")

            if state == "Running":
                # If the instance is already running, skip starting it
                logger.info(
                    f"Instance {path_to_vm} is already running. Skipping start."
                )
                return

            if state == "Stopped":
                # Start the instance if it's currently stopped
                req = ecs_models.StartInstanceRequest(instance_id=path_to_vm)
                self.client.start_instance(req)
                logger.info(f"Instance {path_to_vm} is starting...")

                # Wait until the instance reaches 'Running' state
                _wait_for_instance_running(self.client, path_to_vm)
                logger.info(f"Instance {path_to_vm} is now running.")
            else:
                # For all other states (Pending, Starting, etc.), log a warning
                logger.warning(
                    f"Instance {path_to_vm} is in state '{state}' and cannot be started."
                )

        except Exception as e:
            logger.error(
                f"Failed to start the Aliyun ECS instance {path_to_vm}: {str(e)}"
            )
            raise

    def get_ip_address(self, path_to_vm: str) -> str:
        logger.info("Getting Aliyun ECS instance IP address...")

        try:
            response = self._describe_instance(path_to_vm)
            if not response.body.instances.instance:
                logger.error(f"Instance {path_to_vm} not found")
                return ""

            instance = response.body.instances.instance[0]

            # Get private and public IP addresses
            private_ip = ""
            public_ip = ""

            if hasattr(instance, "vpc_attributes") and instance.vpc_attributes:
                private_ip = (
                    instance.vpc_attributes.private_ip_address.ip_address[0]
                    if instance.vpc_attributes.private_ip_address.ip_address
                    else ""
                )

            if hasattr(instance, "public_ip_address") and instance.public_ip_address:
                public_ip = (
                    instance.public_ip_address.ip_address[0]
                    if instance.public_ip_address.ip_address
                    else ""
                )

            if hasattr(instance, "eip_address") and instance.eip_address:
                public_ip = instance.eip_address.ip_address or public_ip

            # Select which IP to use based on configuration
            ip_to_use = private_ip if (self.use_private_ip and private_ip) else public_ip

            if not ip_to_use:
                logger.warning("No usable IP address available (private/public both missing)")
                return ""

            _wait_until_server_ready(ip_to_use)
            if public_ip:
                vnc_url = f"http://{public_ip}:5910/vnc.html"
                logger.info(f"🖥️  VNC Web Access URL: {vnc_url}")
                logger.info("=" * 80)
            logger.info(f"📡 Public IP: {public_ip}")
            logger.info(f"🏠 Private IP: {private_ip}")
            logger.info(f"🔧 Using IP: {'Private' if ip_to_use == private_ip else 'Public'} -> {ip_to_use}")
            logger.info("=" * 80)
            print(f"\n🌐 VNC Web Access URL: {vnc_url}")
            print(
                "📍 Please open the above address in the browser "
                "for remote desktop access\n"
            )

            return ip_to_use

        except Exception as e:
            logger.error(
                f"Failed to retrieve IP address for the instance {path_to_vm}: {str(e)}"
            )
            raise

    def save_state(self, path_to_vm: str, snapshot_name: str):
        logger.info("Saving Aliyun ECS instance state...")

        try:
            req = ecs_models.CreateImageRequest(
                region_id=self.region,
                instance_id=path_to_vm,
                image_name=snapshot_name,
                description=f"Snapshot created at {datetime.now().isoformat()}",
            )
            response = self.client.create_image(req)
            image_id = response.body.image_id
            logger.info(
                f"Image {image_id} created successfully from instance {path_to_vm}."
            )
            return image_id

        except Exception as e:
            logger.error(
                f"Failed to create image from the instance {path_to_vm}: {str(e)}"
            )
            raise

    def revert_to_snapshot(self, path_to_vm: str, snapshot_name: str):
        logger.info(
            f"Reverting Aliyun ECS instance to snapshot image: {snapshot_name}..."
        )

        try:
            # Step 1: Retrieve the original instance details
            response = self._describe_instance(path_to_vm)
            if not response.body.instances.instance:
                logger.error(f"Instance {path_to_vm} not found")
                return
            # Step 2: Delete the old instance
            req = ecs_models.DeleteInstancesRequest(
                region_id=self.region, instance_id=[path_to_vm], force=True
            )
            self.client.delete_instances(req)
            logger.info(f"Old instance {path_to_vm} has been deleted.")

            # Step 3: Launch a new instance from the snapshot image
            new_instance_id = _allocate_vm()
            logger.info(f"Instance {new_instance_id} is ready.")

            # Get VNC access information
            self.get_ip_address(new_instance_id)

            return new_instance_id

        except Exception as e:
            logger.error(
                f"Failed to revert to snapshot {snapshot_name} for the instance {path_to_vm}: {str(e)}"
            )
            raise

    def stop_emulator(self, path_to_vm: str, region: str = None):
        logger.info(f"Stopping Aliyun ECS instance {path_to_vm}...")

        try:
            req = ecs_models.DeleteInstancesRequest(
                region_id=self.region, instance_id=[path_to_vm], force=True
            )
            self.client.delete_instances(req)
            logger.info(f"Instance {path_to_vm} has been deleted.")

        except Exception as e:
            logger.error(
                f"Failed to stop the Aliyun ECS instance {path_to_vm}: {str(e)}"
            )
            raise

    def _describe_instance(
        self, instance_id: str
    ) -> ecs_models.DescribeInstancesResponse:
        """Get instance details"""
        req = ecs_models.DescribeInstancesRequest(
            region_id=self.region, instance_ids=UtilClient.to_jsonstring([instance_id])
        )
        return self.client.describe_instances(req)



# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0
#
# Portions derived from  https://github.com/microsoft/autogen are under the MIT License.
# SPDX-License-Identifier: MIT
from ..cache.cache import Cache
from .anthropic import AnthropicLLMConfigEntry
from .bedrock import BedrockLLMConfigEntry
from .cerebras import CerebrasLLMConfigEntry
from .client import AzureOpenAILLMConfigEntry, DeepSeekLLMConfigEntry, ModelClient, OpenAILLMConfigEntry, OpenAIWrapper
from .cohere import CohereLLMConfigEntry
from .gemini import GeminiLLMConfigEntry
from .groq import GroqLLMConfigEntry
from .mistral import MistralLLMConfigEntry
from .ollama import OllamaLLMConfigEntry
from .openai_utils import (
    config_list_from_dotenv,
    config_list_from_json,
    config_list_from_models,
    config_list_gpt4_gpt35,
    config_list_openai_aoai,
    filter_config,
    get_config_list,
    get_first_llm_config,
)
from .together import TogetherLLMConfigEntry

__all__ = [
    "AnthropicLLMConfigEntry",
    "AzureOpenAILLMConfigEntry",
    "BedrockLLMConfigEntry",
    "Cache",
    "CerebrasLLMConfigEntry",
    "CohereLLMConfigEntry",
    "DeepSeekLLMConfigEntry",
    "GeminiLLMConfigEntry",
    "GroqLLMConfigEntry",
    "MistralLLMConfigEntry",
    "ModelClient",
    "OllamaLLMConfigEntry",
    "OpenAILLMConfigEntry",
    "OpenAIWrapper",
    "TogetherLLMConfigEntry",
    "config_list_from_dotenv",
    "config_list_from_json",
    "config_list_from_models",
    "config_list_gpt4_gpt35",
    "config_list_openai_aoai",
    "filter_config",
    "get_config_list",
    "get_first_llm_config",
]



import base64
import json
import logging
import time
import os
from io import BytesIO
from typing import Dict, List, Tuple

import backoff
import openai
from PIL import Image
from requests.exceptions import SSLError
from google.api_core.exceptions import (
    InvalidArgument,
    ResourceExhausted,
    InternalServerError,
    BadRequest,
)
from mm_agents.utils.qwen_vl_utils import smart_resize



logger = None

MAX_RETRY_TIMES = 5

def encode_image(image_content):
    return base64.b64encode(image_content).decode("utf-8")


def process_image(image_bytes):
    """
    Process an image for Qwen VL models.
    Resize the image to dimensions expected by the model.
    
    Args:
        image_bytes: Raw image bytes
    
    Returns:
        Base64 encoded image string of the processed image
    """
    # Open image from bytes
    image = Image.open(BytesIO(image_bytes))
    width, height = image.size
    
    # Calculate resized dimensions
    resized_height, resized_width = smart_resize(
        height=height, 
        width=width
    )
    
    # Resize the image
    image = image.resize((resized_width, resized_height))
    
    # Convert to bytes
    buffer = BytesIO()
    image.save(buffer, format="PNG")
    processed_bytes = buffer.getvalue()
    
    # Return base64 encoded string
    return base64.b64encode(processed_bytes).decode('utf-8')


class Qwen25VLAgent:

    def __init__(
        self,
        platform="ubuntu",
        model="qwen2.5-vl-72b-instruct",
        max_tokens=1500,
        top_p=0.9,
        temperature=0.5,
        action_space="pyautogui",
        observation_type="screenshot",
        history_n=4,  # Number of previous interactions to include in full detail
        add_thought_prefix=False,
    ):
        self.platform = platform
        self.model = model
        self.max_tokens = max_tokens
        self.top_p = top_p
        self.temperature = temperature
        self.action_space = action_space
        self.observation_type = observation_type
        self.history_n = history_n  # Control how many previous interactions to include
        self.add_thought_prefix = add_thought_prefix
        assert action_space in ["pyautogui"], "Invalid action space"
        assert observation_type in ["screenshot"], "Invalid observation type"
        self.thoughts = []
        self.actions = []
        self.observations = []
        self.responses = []  # Store model responses
        self.screenshots = []  # Store processed screenshots

    def predict(self, instruction: str, obs: Dict) -> List:
        """
        Predict the next action(s) based on the current observation.
        """
        # Process the screenshot image
        screenshot_bytes = obs["screenshot"]
        
        # Display original dimensions
        image = Image.open(BytesIO(screenshot_bytes))
        width, height = image.size
        print(f"Original screen resolution: {width}x{height}")
        
        # Process the image
        processed_image = process_image(screenshot_bytes)
        processed_img = Image.open(BytesIO(base64.b64decode(processed_image)))
        processed_width, processed_height = processed_img.size
        print(f"Processed image resolution: {processed_width}x{processed_height}")
        
        # Save the current screenshot to history
        self.screenshots.append(processed_image)
        
        # Calculate history window start index
        current_step = len(self.actions)
        history_start_idx = max(0, current_step - self.history_n)
        
        # Build previous actions string - only include actions outside the history window
        previous_actions = []
        for i in range(history_start_idx):
            if i < len(self.actions):
                previous_actions.append(f"Step {i+1}: {self.actions[i]}")
        previous_actions_str = "\n".join(previous_actions) if previous_actions else "None"

        # System prompt with tool definition
        tools_def = {
            "type": "function", 
            "function": {
                "name_for_human": "computer_use", 
                "name": "computer_use", 
                "description": "Use a mouse and keyboard to interact with a computer, and take screenshots.",
                "parameters": {
                    "properties": {
                        "action": {
                            "description": "The action to perform.",
                            "enum": ["key", "type", "mouse_move", "left_click", "left_click_drag", 
                                     "right_click", "middle_click", "double_click", "scroll", "wait", "terminate"], 
                            "type": "string"
                        },
                        "keys": {"description": "Required only by `action=key`.", "type": "array"}, 
                        "text": {"description": "Required only by `action=type`.", "type": "string"}, 
                        "coordinate": {"description": "The x,y coordinates for mouse actions.", "type": "array"}, 
                        "pixels": {"description": "The amount of scrolling.", "type": "number"}, 
                        "time": {"description": "The seconds to wait.", "type": "number"}, 
                        "status": {
                            "description": "The status of the task.", 
                            "type": "string", 
                            "enum": ["success", "failure"]
                        }
                    }, 
                    "required": ["action"], 
                    "type": "object"
                }, 
                "args_format": "Format the arguments as a JSON object."
            }
        }
        
        system_prompt = """You are a helpful assistant

# Tools

You may call one or more functions to assist with the user query.

You are provided with function signatures within <tools></tools> XML tags:
<tools>
""" + json.dumps(tools_def) + """
</tools>

For each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:
<tool_call>
{"name": <function-name>, "arguments": <args-json-object>}
</tool_call>"""

        # Create instruction prompt
        instruction_prompt = f"""
Please generate the next move according to the UI screenshot, instruction and previous actions.

Instruction: {instruction}

Previous actions:
{previous_actions_str}"""

        # Initialize messages with system prompt
        messages = [
            {
                "role": "system",
                "content": [{
                    "type": "text",
                    "text": system_prompt
                }]
            }
        ]
        
        # Add history responses and images within the history window
        history_len = min(self.history_n, len(self.responses))
        if history_len > 0:
            # Only include the most recent history_n steps
            history_responses = self.responses[-history_len:]
            history_screenshots = self.screenshots[-history_len-1:-1]  # Include one more for the previous screenshot
            
            # Add history in conversation format
            for idx in range(history_len):
                # Add the screenshot (user message)
                if idx < len(history_screenshots):
                    screenshot_b64 = history_screenshots[idx]
                    
                    # If this is the first history item, include instruction prompt
                    if idx == 0:
                        messages.append({
                            "role": "user",
                            "content": [
                                {
                                    "type": "image_url",
                                    "image_url": {
                                        "url": f"data:image/png;base64,{screenshot_b64}"
                                    }
                                },
                                {
                                    "type": "text",
                                    "text": instruction_prompt
                                }
                            ]
                        })
                    else:
                        messages.append({
                            "role": "user",
                            "content": [
                                {
                                    "type": "image_url",
                                    "image_url": {
                                        "url": f"data:image/png;base64,{screenshot_b64}"
                                    }
                                }
                            ]
                        })
                
                # Add the action and response (assistant message)
                
                messages.append({
                    "role": "assistant",
                    "content": [
                        {"type": "text", "text": history_responses[idx]}
                    ]
                })
            
            # Add the current screenshot without instruction (since we already have history)
            messages.append({
                "role": "user",
                "content": [
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/png;base64,{processed_image}"
                        }
                    }
                ]
            })
        else:
            # If no history, just add current screenshot with instruction prompt
            messages.append({
                "role": "user",
                "content": [
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/png;base64,{processed_image}"
                        }
                    },
                    {
                        "type": "text",
                        "text": instruction_prompt
                    }
                ]
            })

        # append_text = f"""Step {current_step+1}: Thought:"""
        if self.add_thought_prefix:
            append_text = f"""Thought:"""
            messages.append({"role": "assistant", "content": [{"type": "text", "text": append_text}]})

        # Call the LLM
        response = self.call_llm(
            {
                "model": self.model,
                "messages": messages,
                "max_tokens": self.max_tokens,
                "top_p": self.top_p,
                "temperature": self.temperature,
            },
            self.model,
        )

        logger.info(f"Qwen25VL Output: {response}")
        
        # Save response to history
        self.responses.append(response)

        # Parse response and extract pyautogui code
        low_level_instruction, pyautogui_code = self.parse_response(
            response, 
            width, 
            height, 
            processed_width, 
            processed_height
        )

        logger.info(f"Low level instruction: {low_level_instruction}")
        logger.info(f"Pyautogui code: {pyautogui_code}")

        # Add the action to history
        self.actions.append(low_level_instruction)

        return response, pyautogui_code

    def parse_response(self, response: str, original_width: int = None, original_height: int = None, 
                       processed_width: int = None, processed_height: int = None) -> Tuple[str, List[str]]:
        """
        Parse LLM response and convert it to low level action and pyautogui code.
        
        Args:
            response: Raw response string from the model
            original_width: Width of the original screenshot
            original_height: Height of the original screenshot
            processed_width: Width of the processed image
            processed_height: Height of the processed image
            
        Returns:
            Tuple of (low_level_instruction, list of pyautogui_commands)
        """
        low_level_instruction = ""
        pyautogui_code = []
        
        if response is None or not response.strip():
            return low_level_instruction, pyautogui_code
        
        # Define function to adjust coordinates based on original and processed dimensions
        def adjust_coordinates(x: float, y: float) -> Tuple[int, int]:
            """
            Adjust coordinates from processed image dimensions to original image dimensions.
            """
            if all([original_width, original_height, processed_width, processed_height]):
                # Calculate the scale factors between original and processed images
                x_scale = original_width / processed_width
                y_scale = original_height / processed_height
                
                # Apply scaling to get coordinates in original image space
                adjusted_x = int(x * x_scale)
                adjusted_y = int(y * y_scale)
                
                return adjusted_x, adjusted_y
            else:
                # If any dimension is missing, return the original coordinates
                return int(x), int(y)
        
        # Define inner function to process tool calls
        def process_tool_call(json_str: str) -> None:
            """Process a single tool call JSON string."""
            try:
                # Parse the JSON
                tool_call = json.loads(json_str)
                if tool_call.get("name") == "computer_use":
                    # Convert computer_use actions to pyautogui commands
                    args = tool_call["arguments"]
                    action = args["action"]
                    
                    if action == "left_click":
                        if "coordinate" in args:
                            x, y = args["coordinate"]
                            adj_x, adj_y = adjust_coordinates(x, y)
                            pyautogui_code.append(f"pyautogui.click({adj_x}, {adj_y})")
                        else:
                            pyautogui_code.append("pyautogui.click()")
                            
                    elif action == "right_click":
                        if "coordinate" in args:
                            x, y = args["coordinate"]
                            adj_x, adj_y = adjust_coordinates(x, y)
                            pyautogui_code.append(f"pyautogui.rightClick({adj_x}, {adj_y})")
                        else:
                            pyautogui_code.append("pyautogui.rightClick()")
                            
                    elif action == "middle_click":
                        if "coordinate" in args:
                            x, y = args["coordinate"]
                            adj_x, adj_y = adjust_coordinates(x, y)
                            pyautogui_code.append(f"pyautogui.middleClick({adj_x}, {adj_y})")
                        else:
                            pyautogui_code.append("pyautogui.middleClick()")
                            
                    elif action == "double_click":
                        if "coordinate" in args:
                            x, y = args["coordinate"]
                            adj_x, adj_y = adjust_coordinates(x, y)
                            pyautogui_code.append(f"pyautogui.doubleClick({adj_x}, {adj_y})")
                        else:
                            pyautogui_code.append("pyautogui.doubleClick()")
                            
                    elif action == "type":
                        text = args.get("text", "")
                        pyautogui_code.append(f"pyautogui.typewrite('{text}')")
                        
                    elif action == "key":
                        keys = args.get("keys", [])
                        # Fix possible formatting issues in the keys parameter
                        if isinstance(keys, list):
                            # Clean up any formatting issues in the keys
                            cleaned_keys = []
                            for key in keys:
                                # Check if the key has the "keys=[" prefix or "]" suffix
                                if isinstance(key, str):
                                    # Remove "keys=[" prefix if present
                                    if key.startswith("keys=["):
                                        key = key[6:]
                                    # Remove "]" suffix if present
                                    if key.endswith("]"):
                                        key = key[:-1]
                                    # Handle case where string contains representation of list items
                                    if key.startswith("['") or key.startswith("[\""):
                                        key = key[2:] if len(key) > 2 else key
                                    if key.endswith("']") or key.endswith("\"]"):
                                        key = key[:-2] if len(key) > 2 else key
                                    # Strip any extra whitespace
                                    key = key.strip()
                                    # Add to cleaned keys
                                    cleaned_keys.append(key)
                                else:
                                    cleaned_keys.append(key)
                            keys = cleaned_keys
                            
                        # Format the keys for hotkey or press command
                        keys_str = ", ".join([f"'{key}'" for key in keys])
                        if len(keys) > 1:
                            pyautogui_code.append(f"pyautogui.hotkey({keys_str})")
                        else:
                            pyautogui_code.append(f"pyautogui.press({keys_str})")
                            
                    elif action == "scroll":
                        pixels = args.get("pixels", 0)
                        pyautogui_code.append(f"pyautogui.scroll({pixels})")
                        
                    elif action == "wait":
                        pyautogui_code.append("WAIT")  # Special code for wait action
                        
                    elif action == "terminate":
                        pyautogui_code.append("DONE")  # Special code for termination
                        
                    elif action == "mouse_move":
                        if "coordinate" in args:
                            x, y = args["coordinate"]
                            adj_x, adj_y = adjust_coordinates(x, y)
                            pyautogui_code.append(f"pyautogui.moveTo({adj_x}, {adj_y})")
                        else:
                            pyautogui_code.append("pyautogui.moveTo(0, 0)")
                            
                    elif action == "left_click_drag":
                        if "coordinate" in args:
                            x, y = args["coordinate"]
                            adj_x, adj_y = adjust_coordinates(x, y)
                            duration = args.get("duration", 0.5)
                            pyautogui_code.append(f"pyautogui.dragTo({adj_x}, {adj_y}, duration={duration})")
                        else:
                            pyautogui_code.append("pyautogui.dragTo(0, 0)")
            except (json.JSONDecodeError, KeyError) as e:
                logger.error(f"Failed to parse tool call: {e}")
        
        # Parse the response line by line
        lines = response.split('\n')
        inside_tool_call = False
        current_tool_call = []
        
        for line in lines:
            line = line.strip()
            if not line:
                continue
            
            # Extract low-level instruction from lines starting with "Action:" or similar
            if line.lower().startswith(("action:", "step", "i will", "i'll", "now i")):
                if not low_level_instruction:
                    # Only store the first action description as low level instruction
                    low_level_instruction = line
                continue
            
            # Handle lines inside tool call markers
            if line.startswith("<tool_call>") or line.startswith("⚗") or line.startswith("📐"): # Yeah, it's a bug during data processing
                inside_tool_call = True
                continue
            elif line.startswith("</tool_call>") or line.startswith("⚗") or line.startswith("📐"): # Yeah, it's a bug during data processing
                if current_tool_call:
                    # Process the collected tool call
                    process_tool_call("\n".join(current_tool_call))
                    current_tool_call = []
                inside_tool_call = False
                continue
            
            if inside_tool_call:
                current_tool_call.append(line)
                continue
            
            # Try to parse individual lines as JSON
            if line.startswith("{") and line.endswith("}"):
                try:
                    json_obj = json.loads(line)
                    if "name" in json_obj and "arguments" in json_obj:
                        process_tool_call(line)
                except json.JSONDecodeError:
                    pass
        
        # Process any remaining tool call content
        if current_tool_call:
            process_tool_call("\n".join(current_tool_call))
        
        # If we still don't have a low-level instruction, generate a default one
        if not low_level_instruction and len(pyautogui_code) > 0:
            action_type = pyautogui_code[0].split(".", 1)[1].split("(", 1)[0]
            low_level_instruction = f"Performing {action_type} action"
        
        return low_level_instruction, pyautogui_code

    @backoff.on_exception(
        backoff.constant,
        # here you should add more model exceptions as you want,
        # but you are forbidden to add "Exception", that is, a common type of exception
        # because we want to catch this kind of Exception in the outside to ensure
        # each example won't exceed the time limit
        (
            # General exceptions
            SSLError,
            # OpenAI exceptions
            openai.RateLimitError,
            openai.BadRequestError,
            openai.InternalServerError,
            # Google exceptions
            InvalidArgument,
            ResourceExhausted,
            InternalServerError,
            BadRequest,
            # Groq exceptions
            # todo: check
        ),
        interval=30,
        max_tries=5,
    )
    def call_llm(self, payload, model):
        messages = payload["messages"]

        base_url = os.getenv('DASHSCOPE_BASE_URL', "https://dashscope.aliyuncs.com/compatible-mode/v1")
        api_key = os.getenv('DASHSCOPE_API_KEY', "sk-123")
        
        client = openai.OpenAI(
            base_url=base_url,
            api_key=api_key
        )

        for _ in range(MAX_RETRY_TIMES):
            logger.info("Generating content with Qwen model: %s", model)
            try:
                response = client.chat.completions.create(
                    model=model,
                    messages=messages,
                    max_tokens=self.max_tokens,
                    temperature=self.temperature,
                    top_p=self.top_p
                )
                return response.choices[0].message.content
            except Exception as e:
                logger.error(f"Error calling Qwen model: {e}")
                time.sleep(5)
                continue
        return ""

    def reset(self, _logger=None):
        global logger
        logger = (_logger if _logger is not None else
                  logging.getLogger("desktopenv.qwen25vl_agent"))

        self.thoughts = []
        self.action_descriptions = []
        self.actions = []
        self.observations = []
        self.responses = []  # Reset responses
        self.screenshots = []  # Reset screenshots



# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0
#
# Portions derived from  https://github.com/microsoft/autogen are under the MIT License.
# SPDX-License-Identifier: MIT

from typing import Any, Literal, TypedDict, Union

MessageContentType = Union[str, list[Union[dict[str, Any], str]], None]


class UserMessageTextContentPart(TypedDict):
    """Represents a text content part of a user message"""

    type: Literal["text"]
    """The type of the content part. Always "text" for text content parts."""
    text: str
    """The text content of the part."""


class UserMessageImageContentPart(TypedDict):
    """Represents an image content part of a user message"""

    type: Literal["input_image"]
    """The type of the content part. Always "input_image" for image content parts."""
    # Ignoring the other "detail param for now"
    image_url: str
    """The URL of the image."""



#!/usr/bin/env python3
# python gui_agents/maestro/debug_system/run_from_snapshot.py
# python -m gui_agents.maestro.debug_system.run_from_snapshot
"""
Debug helper: restore MainController from a snapshot and optionally run it.
"""


import argparse
import logging
import time
from typing import Optional, List, Any

from ..data_models import CommandData
from ..new_global_state import NewGlobalState
from ..snapshot_restorer import (
    restore_maincontroller_from_globalstate,
)
from .logging_setup import setup_file_logging

logger = logging.getLogger(__name__)


def _extract_commands_sorted(controller) -> List[CommandData]:
    """Get all commands (CommandData) and sort them in ascending order by created_at."""
    gs: NewGlobalState = getattr(controller, "global_state")
    if gs is None or not hasattr(gs, "get_commands"):
        return []
    try:
        cmds: List[CommandData] = gs.get_commands() or []
    except Exception:
        return []
    try:
        return sorted(cmds, key=lambda c: getattr(c, "created_at", 0))
    except Exception:
        return cmds


def run_main_controller_from_snapshot(
    runtime_dir: str,
    snapshot_id: Optional[str] = None,
    target_dir: Optional[str] = None,
    os_word_task_id: Optional[str] = None,
):
    """
    Restore a MainController from an existing snapshot and optionally run it.

    Args:
        runtime_dir: Path to the runtime directory (e.g., "runtime/20250826_141730")
        snapshot_id: Snapshot ID (e.g., "snapshot_20250826_141736"); if None, interactive selection
        target_dir: Restore target directory; if None, it will be generated automatically
        auto_run: Whether to immediately execute the main loop

    Returns:
        The restored MainController if successful; otherwise None.
    """
    result = restore_maincontroller_from_globalstate(
        runtime_dir=runtime_dir,
        snapshot_id=snapshot_id,
        target_dir=target_dir,
        os_word_task_id=os_word_task_id
    )
    if result is None:
        logger.error("Failed to restore MainController from snapshot")
        return None

    controller, target_path, config_params = result

    # Write file logs to target_path
    setup_file_logging(target_path)

    logger.info(f"MainController restored from snapshot successfully. Logs at: {target_path}")

    # Execute commands in chronological order before running the main loop, with 5-second intervals
    try:
        commands = _extract_commands_sorted(controller)
        if commands:
            logger.info(f"Pre-executing {len(commands)} commands before main loop...")
        for idx, cmd in enumerate(commands, 1):
            try:
                logger.info(f"[PreExec {idx}/{len(commands)}] Executing command_id={getattr(cmd, 'command_id', 'N/A')}")
                executor = getattr(controller, "executor", None)
                if executor is not None and hasattr(executor, "execute_command"):
                    executor.execute_command(cmd)  # Execute only, do not touch global_state
                else:
                    logger.warning("Controller has no executor.execute_command; skipping pre-exec")
                    break
            except Exception as e:
                logger.warning(f"Pre-exec command failed: {e}")
            time.sleep(5)
    except Exception as e:
        logger.warning(f"Failed to pre-execute commands: {e}")

    controller.execute_main_loop()

    return controller


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Run MainController from snapshot")
    parser.add_argument("runtime_dir", type=str, help="Path to runtime directory")
    parser.add_argument("--snapshot", "-s", type=str, default=None, help="Snapshot ID")
    parser.add_argument("--target", "-t", type=str, default=None, help="Restore target directory")
    parser.add_argument("--no-run", action="store_true", help="Do not auto-run main loop")
    return parser.parse_args()


if __name__ == "__main__":
    import logging
    from .logging_setup import setup_debug_logging, setup_file_logging

    # Simple console logging
    setup_debug_logging(logging.INFO)

    # Write to target_path log file (example with default test target directory)
    # If using CLI arguments, call setup_file_logging(args.target or target_path) after parsing

    # args = _parse_args()

    target_dir=None
    runtime_dir = "runtime/vmrun_20250831_003040/20250831_003117"
    snapshot_id = "snapshot_20250831_004442"
    # runtime_dir = "runtime/20250828_141244"
    # snapshot_id = "snapshot_20250828_141320"
    os_word_task_id = "535364ea-05bd-46ea-9937-9f55c68507e8"

    controller = run_main_controller_from_snapshot(
        runtime_dir=runtime_dir,
        snapshot_id=snapshot_id,
        target_dir=target_dir,
        os_word_task_id=os_word_task_id
    ) 



import os


# Default TTL minutes for instance auto-release (Aliyun-side)
# Can be overridden via environment variable DEFAULT_TTL_MINUTES
# ATTENTION: ECS requires TTL to be at least 30 minutes (if TTL > 0)
MIN_TTL_MINUTES: int = 30

_ttl_env_str = os.getenv("DEFAULT_TTL_MINUTES", "60")
try:
    _ttl_env_val = int(_ttl_env_str)
except Exception:
    _ttl_env_val = 60

# If TTL is positive but less than Aliyun minimum, clamp to 30 minutes
if _ttl_env_val > 0 and _ttl_env_val < MIN_TTL_MINUTES:
    DEFAULT_TTL_MINUTES: int = MIN_TTL_MINUTES
else:
    DEFAULT_TTL_MINUTES: int = _ttl_env_val

# Master switch for TTL feature
ENABLE_TTL: bool = os.getenv("ENABLE_TTL", "true").lower() == "true"


def compute_ttl_seconds(ttl_minutes: int) -> int:
    try:
        return max(0, int(ttl_minutes) * 60)
    except Exception:
        return 0





# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0


import copy
import logging
import os
from typing import Annotated, Any, Literal, Optional, Type, Union

from pydantic import BaseModel

from ....doc_utils import export_module
from ....import_utils import optional_import_block, require_optional_import
from ....llm_config import LLMConfig
from ... import Tool

with optional_import_block():
    from openai import OpenAI
    from openai.types.responses import WebSearchToolParam
    from openai.types.responses.web_search_tool import UserLocation


@require_optional_import("openai>=1.66.2", "openai")
@export_module("autogen.tools.experimental")
class WebSearchPreviewTool(Tool):
    """WebSearchPreviewTool is a tool that uses OpenAI's web_search_preview tool to perform a search."""

    def __init__(
        self,
        *,
        llm_config: Union[LLMConfig, dict[str, Any]],
        search_context_size: Literal["low", "medium", "high"] = "medium",
        user_location: Optional[dict[str, str]] = None,
        instructions: Optional[str] = None,
        text_format: Optional[Type[BaseModel]] = None,
    ):
        """Initialize the WebSearchPreviewTool.

        Args:
            llm_config: The LLM configuration to use. This should be a dictionary
                containing the model name and other parameters.
            search_context_size: The size of the search context. One of `low`, `medium`, or `high`.
                `medium` is the default.
            user_location: The location of the user. This should be a dictionary containing
                the city, country, region, and timezone.
            instructions: Inserts a system (or developer) message as the first item in the model's context.
            text_format: The format of the text to be returned. This should be a subclass of `BaseModel`.
                The default is `None`, which means the text will be returned as a string.
        """
        self.web_search_tool_param = WebSearchToolParam(
            type="web_search_preview",
            search_context_size=search_context_size,
            user_location=UserLocation(**user_location) if user_location else None,  # type: ignore[typeddict-item]
        )
        self.instructions = instructions
        self.text_format = text_format

        if isinstance(llm_config, LLMConfig):
            llm_config = llm_config.model_dump()

        llm_config = copy.deepcopy(llm_config)

        if "config_list" not in llm_config:
            raise ValueError("llm_config must contain 'config_list' key")

        # Find first OpenAI model which starts with "gpt-4"
        self.model = None
        self.api_key = None
        for model in llm_config["config_list"]:
            if model["model"].startswith("gpt-4") and model.get("api_type", "openai") == "openai":
                self.model = model["model"]
                self.api_key = model.get("api_key", os.getenv("OPENAI_API_KEY"))
                break
        if self.model is None:
            raise ValueError(
                "No OpenAI model starting with 'gpt-4' found in llm_config, other models do not support web_search_preview"
            )

        if not self.model.startswith("gpt-4.1") and not self.model.startswith("gpt-4o-search-preview"):
            logging.warning(
                f"We recommend using a model starting with 'gpt-4.1' or 'gpt-4o-search-preview' for web_search_preview, but found {self.model}. "
                "This may result in suboptimal performance."
            )

        def web_search_preview(
            query: Annotated[str, "The search query. Add all relevant context to the query."],
        ) -> Union[str, Optional[BaseModel]]:
            client = OpenAI()

            if not self.text_format:
                response = client.responses.create(
                    model=self.model,  # type: ignore[arg-type]
                    tools=[self.web_search_tool_param],
                    input=query,
                    instructions=self.instructions,
                )
                return response.output_text

            else:
                response = client.responses.parse(
                    model=self.model,  # type: ignore[arg-type]
                    tools=[self.web_search_tool_param],
                    input=query,
                    instructions=self.instructions,
                    text_format=self.text_format,
                )
                return response.output_parsed

        super().__init__(
            name="web_search_preview",
            description="Tool used to perform a web search. It can be used as google search or directly searching a specific website.",
            func_or_tool=web_search_preview,
        )



# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0

from .base_client import OpenAIRealtimeClient
from .rtc_client import OpenAIRealtimeWebRTCClient

__all__ = ["OpenAIRealtimeClient", "OpenAIRealtimeWebRTCClient"]



"""
New Manager Module for GUI-Agent Architecture
Responsible for task planning, decomposition, and resource allocation
"""

import logging
import time
from datetime import datetime
from typing import Dict, Any, Union

from ..tools.new_tools import NewTools
from ..core.new_knowledge import NewKnowledgeBase

from .new_global_state import NewGlobalState
from .enums import ManagerStatus
from .manager.plan import PlanningHandler, PlanningScenario, PlanningResult
from .manager.supplement import SupplementHandler

logger = logging.getLogger(__name__)


class NewManager:
    """
    Enhanced Manager module for GUI-Agent architecture
    Responsible for task planning, decomposition, and resource allocation

    Note: Planning prompts include generic guidance for configuration persistence (prefer editing config files; GUI changes require Save/Exit) and role assignment (Technician vs Operator). See planning_helpers.generate_planning_prompt.
    """

    def __init__(
        self,
        tools_dict: Dict[str, Any],
        global_state: NewGlobalState,
        local_kb_path: str = "",
        platform: str = "Windows",
        enable_search: bool = False,
        enable_narrative: bool = False,
        max_replan_attempts: int = 3,
    ):
        """
        Initialize the Manager module
        
        Args:
            tools_dict: Dictionary containing tool configurations
            global_state: Global state instance
            local_kb_path: Path to local knowledge base
            platform: Target platform (Windows/Darwin/Ubuntu)
            enable_search: Whether to enable web search
            max_replan_attempts: Maximum replanning attempts
        """
        self.tools_dict = tools_dict
        self.global_state = global_state
        self.local_kb_path = local_kb_path
        self.platform = platform
        self.enable_search = enable_search
        self.enable_narrative = enable_narrative
        self.max_replan_attempts = max_replan_attempts

        # Initialize status
        self.status = ManagerStatus.IDLE
        self.plan_scenario = PlanningScenario.REPLAN

        # Initialize tools
        self._initialize_tools()

        # Initialize knowledge base
        self._initialize_knowledge_base()

        # Initialize handlers
        self._initialize_handlers()

        logger.info("NewManager initialized successfully")

    def _initialize_tools(self):
        """Initialize required tools with backward-compatible keys"""
        self.planner_agent_name = "planner_role"
        self.supplement_agent_name = "supplement_role"
        self.dag_translator_agent_name = "dag_translator"

        # planner_agent
        self.planner_agent = NewTools()
        self.planner_agent.register_tool(
            self.planner_agent_name,
            self.tools_dict[self.planner_agent_name]["provider"],
            self.tools_dict[self.planner_agent_name]["model"],
        )

        # dag_translator_agent
        self.dag_translator_agent = NewTools()
        self.dag_translator_agent.register_tool(
            self.dag_translator_agent_name,
            self.tools_dict[self.dag_translator_agent_name]["provider"],
            self.tools_dict[self.dag_translator_agent_name]["model"],
        )

        # supplement_agent
        self.supplement_agent = NewTools()
        self.supplement_agent.register_tool(
            self.supplement_agent_name,
            self.tools_dict[self.supplement_agent_name]["provider"],
            self.tools_dict[self.supplement_agent_name]["model"]
        )

        # objective_alignment agent
        self.objective_alignment_agent = None
        if self.tools_dict.get("objective_alignment"):
            try:
                self.objective_alignment_agent = NewTools()
                self.objective_alignment_agent.register_tool(
                    "objective_alignment",
                    self.tools_dict["objective_alignment"]["provider"],
                    self.tools_dict["objective_alignment"]["model"],
                )
            except Exception:
                self.objective_alignment_agent = None

        # Embedding engine for Memory
        self.embedding_engine = NewTools()
        self.embedding_engine.register_tool(
            "embedding",
            self.tools_dict["embedding"]["provider"],
            self.tools_dict["embedding"]["model"],
        )

        # Web search engine (optional)
        if self.enable_search and self.tools_dict.get("websearch"):
            self.search_engine = NewTools()
            self.search_engine.register_tool(
                "websearch",
                self.tools_dict["websearch"]["provider"],
                self.tools_dict["websearch"]["model"],
            )
        else:
            self.search_engine = None

    def _initialize_knowledge_base(self):
        """Initialize knowledge base for RAG operations"""
        kb_tools_dict = {
            "query_formulator": self.tools_dict.get("query_formulator", {}),
            "context_fusion": self.tools_dict.get("context_fusion", {}),
            "narrative_summarization": self.tools_dict.get("narrative_summarization", {}),
            "episode_summarization": self.tools_dict.get("episode_summarization", {}),
        }

        self.knowledge_base = NewKnowledgeBase(
            embedding_engine=self.embedding_engine,
            local_kb_path=self.local_kb_path,
            platform=self.platform,
            Tools_dict=kb_tools_dict,
        )

    def _initialize_handlers(self):
        """Initialize the planning and supplement handlers"""
        self.planning_handler = PlanningHandler(
            global_state=self.global_state,
            planner_agent=self.planner_agent,
            dag_translator_agent=self.dag_translator_agent,
            knowledge_base=self.knowledge_base,
            search_engine=self.search_engine,
            platform=self.platform,
            enable_search=self.enable_search,
            enable_narrative=self.enable_narrative,
            objective_alignment_agent=self.objective_alignment_agent,
        )
        
        self.supplement_handler = SupplementHandler(
            global_state=self.global_state,
            supplement_agent=self.supplement_agent,
            search_engine=self.search_engine,
            knowledge_base=self.knowledge_base,
            enable_search=self.tools_dict[self.supplement_agent_name]["enable_search"],
        )

    def plan_task(self, scenario: Union[PlanningScenario, str]) -> PlanningResult:
        """
        Execute task planning based on scenario and current trigger_code
        
        Args:
            scenario: Planning scenario (INITIAL_PLAN|REPLAN|SUPPLEMENT or enum)
            
        Returns:
            PlanningResult: Planning result with subtasks or supplement
        """
        try:
            scenario_enum = self._normalize_scenario(scenario)
            self.status = ManagerStatus.PLANNING
            
            # Get current trigger_code to determine specific planning strategy
            current_trigger_code = self._get_current_trigger_code()
            
            self.global_state.log_operation("manager", "planning_start", {
                "scenario": scenario_enum.value,
                "trigger_code": current_trigger_code,
                "timestamp": time.time()
            })

            if scenario_enum == PlanningScenario.SUPPLEMENT:
                return self._handle_supplement_scenario()
            else:
                return self._handle_planning_scenario(scenario_enum, current_trigger_code)

        except Exception as e:
            logger.error(f"Planning failed: {e}")
            self.status = ManagerStatus.ERROR
            self.global_state.log_operation("manager", "planning_error", {
                "error": str(e),
                "timestamp": time.time()
            })

            return PlanningResult(
                success=False,
                scenario=self._normalize_scenario(scenario).value if isinstance(
                    scenario, str) else scenario.value,
                subtasks=[],
                supplement="",
                reason=f"Planning failed: {str(e)}",
                created_at=datetime.now().isoformat())
        finally:
            self.status = ManagerStatus.IDLE

    def _normalize_scenario(
            self, scenario: Union[PlanningScenario, str]) -> PlanningScenario:
        """Normalize string/enum scenario to PlanningScenario enum (case-insensitive)."""
        if isinstance(scenario, PlanningScenario):
            return scenario
        s = str(scenario).strip().lower()
        if s in {"replan", "re-plan"}:
            return PlanningScenario.REPLAN
        if s in {"supplement", "supp"}:
            return PlanningScenario.SUPPLEMENT
        # Default to INITIAL_PLAN if unknown
        return PlanningScenario.REPLAN

    def _handle_planning_scenario(self, scenario: PlanningScenario, trigger_code: str = "controller") -> PlanningResult:
        """Handle planning scenarios (INITIAL_PLAN/REPLAN) with specific trigger_code context"""
        return self.planning_handler.handle_planning_scenario(scenario, trigger_code)

    def _handle_supplement_scenario(self) -> PlanningResult:
        """Handle supplement collection scenario"""
        try:
            logger.info("Starting supplement scenario handling")
            result = self.supplement_handler.handle_supplement_scenario()
            
            # Validate result structure
            if not isinstance(result, dict):
                logger.error(f"Invalid supplement result type: {type(result)}")
                return PlanningResult(
                    success=False,
                    scenario="supplement",
                    subtasks=[],
                    supplement="",
                    reason="Invalid supplement result structure",
                    created_at=datetime.now().isoformat()
                )
            
            # Log supplement result
            self.global_state.log_operation("manager", "supplement_completed", {
                "success": result.get("success", False),
                "supplement_length": len(result.get("supplement", "")),
                "timestamp": time.time()
            })
            
            return PlanningResult(**result)
            
        except Exception as e:
            logger.error(f"Supplement scenario handling failed: {e}")
            self.global_state.log_operation("manager", "supplement_error", {
                "error": str(e),
                "timestamp": time.time()
            })
            
            return PlanningResult(
                success=False,
                scenario="supplement",
                subtasks=[],
                supplement="",
                reason=f"Supplement handling failed: {str(e)}",
                created_at=datetime.now().isoformat()
            )

    def get_planning_status(self) -> Dict[str, Any]:
        """Get current planning status"""
        return {
            "status": self.status.value,
            "replan_attempts": self.planning_handler.replan_attempts,
            "supplement_attempts": self.supplement_handler.supplement_attempts,
            "planning_history_count": len(self.planning_handler.planning_history),
            "max_replan_attempts": self.max_replan_attempts,
        }

    def reset_planning_state(self):
        """Reset planning state (useful for new tasks)"""
        self.planning_handler.replan_attempts = 0
        self.supplement_handler.supplement_attempts = 0
        self.planning_handler.planning_history = []
        self.status = ManagerStatus.IDLE

        self.global_state.add_event("manager", "planning_reset", "Planning state reset")

    def can_replan(self) -> bool:
        """Check if replanning is still allowed"""
        return self.planning_handler.replan_attempts < self.max_replan_attempts

    def _get_current_trigger_code(self) -> str:
        """Get current trigger_code"""
        controller_state = self.global_state.get_controller_state()
        return controller_state.get("trigger_code", "")


# Export a friendly alias to match the interface name used elsewhere
Manager = NewManager



# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
from collections.abc import AsyncGenerator
from logging import Logger
from typing import Any, AsyncContextManager, Callable, Literal, Optional, Protocol, TypeVar, Union, runtime_checkable

from asyncer import create_task_group

from .....doc_utils import export_module
from .....llm_config import LLMConfig
from ..realtime_events import InputAudioBufferDelta, RealtimeEvent

__all__ = ["RealtimeClientProtocol", "Role", "get_client", "register_realtime_client"]

# define role literal type for typing
Role = Literal["user", "assistant", "system"]


@runtime_checkable
@export_module("autogen.agentchat.realtime.experimental.clients")
class RealtimeClientProtocol(Protocol):
    async def send_function_result(self, call_id: str, result: str) -> None:
        """Send the result of a function call to a Realtime API.

        Args:
            call_id (str): The ID of the function call.
            result (str): The result of the function call.
        """
        ...

    async def send_text(self, *, role: Role, text: str) -> None:
        """Send a text message to a Realtime API.

        Args:
            role (str): The role of the message.
            text (str): The text of the message.
        """
        ...

    async def send_audio(self, audio: str) -> None:
        """Send audio to a Realtime API.

        Args:
            audio (str): The audio to send.
        """
        ...

    async def truncate_audio(self, audio_end_ms: int, content_index: int, item_id: str) -> None:
        """Truncate audio in a Realtime API.

        Args:
            audio_end_ms (int): The end of the audio to truncate.
            content_index (int): The index of the content to truncate.
            item_id (str): The ID of the item to truncate.
        """
        ...

    async def session_update(self, session_options: dict[str, Any]) -> None:
        """Send a session update to a Realtime API.

        Args:
            session_options (dict[str, Any]): The session options to update.
        """
        ...

    def connect(self) -> AsyncContextManager[None]: ...

    def read_events(self) -> AsyncGenerator[RealtimeEvent, None]:
        """Read events from a Realtime Client."""
        ...

    async def _read_from_connection(self) -> AsyncGenerator[RealtimeEvent, None]:
        """Read events from a Realtime connection."""
        ...

    def _parse_message(self, message: dict[str, Any]) -> list[RealtimeEvent]:
        """Parse a message from a Realtime API.

        Args:
            message (dict[str, Any]): The message to parse.

        Returns:
            list[RealtimeEvent]: The parsed events.
        """
        ...

    @classmethod
    def get_factory(
        cls, llm_config: Union[LLMConfig, dict[str, Any]], logger: Logger, **kwargs: Any
    ) -> Optional[Callable[[], "RealtimeClientProtocol"]]:
        """Create a Realtime API client.

        Args:
            llm_config: The config for the client.
            logger: The logger to use for logging events.
            **kwargs: Additional arguments.

        Returns:
            RealtimeClientProtocol: The Realtime API client is returned if the model matches the pattern
        """
        ...


class RealtimeClientBase:
    def __init__(self):
        self._eventQueue = asyncio.Queue()

    async def add_event(self, event: Optional[RealtimeEvent]):
        await self._eventQueue.put(event)

    async def get_event(self) -> Optional[RealtimeEvent]:
        return await self._eventQueue.get()

    async def _read_from_connection_task(self):
        async for event in self._read_from_connection():
            await self.add_event(event)
        await self.add_event(None)

    async def _read_events(self) -> AsyncGenerator[RealtimeEvent, None]:
        """Read events from a Realtime Client."""
        async with create_task_group() as tg:
            tg.start_soon(self._read_from_connection_task)
            while True:
                try:
                    event = await self._eventQueue.get()
                    if event is not None:
                        yield event
                    else:
                        break
                except Exception:
                    break

    async def queue_input_audio_buffer_delta(self, audio: str) -> None:
        """queue InputAudioBufferDelta.

        Args:
            audio (str): The audio.
        """
        await self.add_event(InputAudioBufferDelta(delta=audio, item_id=None, raw_message=dict()))


_realtime_client_classes: dict[str, type[RealtimeClientProtocol]] = {}

T = TypeVar("T", bound=RealtimeClientProtocol)


def register_realtime_client() -> Callable[[type[T]], type[T]]:
    """Register a Realtime API client.

    Returns:
        Callable[[type[T]], type[T]]: The decorator to register the Realtime API client
    """

    def decorator(client_cls: type[T]) -> type[T]:
        """Register a Realtime API client.

        Args:
            client_cls: The client to register.
        """
        global _realtime_client_classes
        fqn = f"{client_cls.__module__}.{client_cls.__name__}"
        _realtime_client_classes[fqn] = client_cls

        return client_cls

    return decorator


@export_module("autogen.agentchat.realtime.experimental.clients")
def get_client(llm_config: Union[LLMConfig, dict[str, Any]], logger: Logger, **kwargs: Any) -> "RealtimeClientProtocol":
    """Get a registered Realtime API client.

    Args:
        llm_config: The config for the client.
        logger: The logger to use for logging events.
        **kwargs: Additional arguments.

    Returns:
        RealtimeClientProtocol: The Realtime API client.
    """
    global _realtime_client_classes
    for _, client_cls in _realtime_client_classes.items():
        factory = client_cls.get_factory(llm_config=llm_config, logger=logger, **kwargs)
        if factory:
            return factory()

    raise ValueError("Realtime API client not found.")



import base64
import json
import logging
import os
import re
import time
from io import BytesIO
from typing import Dict, List

import backoff
import openai
import requests
from PIL import Image
from google.api_core.exceptions import (
    InvalidArgument,
    ResourceExhausted,
    InternalServerError,
    BadRequest,
)
from requests.exceptions import SSLError

logger = None

OPENAI_API_KEY = "Your OpenAI API Key"
JEDI_API_KEY = "Your Jedi API Key"
JEDI_SERVICE_URL = "Your Jedi Service URL"

from mm_agents.prompts import JEDI_PLANNER_SYS_PROMPT, JEDI_GROUNDER_SYS_PROMPT
from mm_agents.utils.qwen_vl_utils import smart_resize

def encode_image(image_content):
    return base64.b64encode(image_content).decode("utf-8")

class JediAgent7B:
    def __init__(
        self,
        platform="ubuntu",
        planner_model="gpt-4o",
        executor_model="jedi-7b",
        max_tokens=1500,
        top_p=0.9,
        temperature=0.5,
        action_space="pyautogui",
        observation_type="screenshot",
        max_steps=15
    ):
        self.platform = platform
        self.planner_model = planner_model
        self.executor_model = executor_model
        assert self.executor_model is not None, "Executor model cannot be None"
        self.max_tokens = max_tokens
        self.top_p = top_p
        self.temperature = temperature
        self.action_space = action_space
        self.observation_type = observation_type
        assert action_space in ["pyautogui"], "Invalid action space"
        assert observation_type in ["screenshot"], "Invalid observation type"
        self.thoughts = []
        self.actions = []
        self.observations = []
        self.observation_captions = []
        self.max_image_history_length = 5
        self.current_step = 1
        self.max_steps = max_steps

    def predict(self, instruction: str, obs: Dict) -> List:
        """
        Predict the next action(s) based on the current observation.
        """

        # get the width and height of the screenshot
        image = Image.open(BytesIO(obs["screenshot"]))
        width, height = image.convert("RGB").size

        previous_actions = ("\n".join([
            f"Step {i+1}: {action}" for i, action in enumerate(self.actions)
        ]) if self.actions else "None")

        user_prompt = (
            f"""Please generate the next move according to the UI screenshot and instruction. And you can refer to the previous actions and observations for reflection.\n\nInstruction: {instruction}\n\n""")

        messages = [{
            "role": "system",
            "content": [{
                "type": "text",
                "text": JEDI_PLANNER_SYS_PROMPT.replace("{current_step}", str(self.current_step)).replace("{max_steps}", str(self.max_steps))
            }]
        }]

        # Determine which observations to include images for (only most recent ones)
        obs_start_idx = max(0, len(self.observations) - self.max_image_history_length)
        
        # Add all thought and action history
        for i in range(len(self.thoughts)):
            # For recent steps, include the actual screenshot
            if i >= obs_start_idx:
                messages.append({
                    "role": "user",
                    "content": [{
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/png;base64,{encode_image(self.observations[i]['screenshot'])}",
                            "detail": "high"
                        },
                    }]
                })
            # For older steps, use the observation caption instead of the image
            else:
                messages.append({
                    "role": "user",
                    "content": [{
                        "type": "text",
                        "text": f"Observation: {self.observation_captions[i]}"
                    }]
                })

            thought_messages = f"Thought:\n{self.thoughts[i]}"

            action_messages = f"Action:"
            for action in self.actions[i]:
                action_messages += f"\n{action}"
            messages.append({
                "role": "assistant",
                "content": [{
                    "type": "text",
                    "text": thought_messages + "\n" + action_messages
                }]
            })
            #print(thought_messages + "\n" + action_messages)

        messages.append({
            "role":"user",
            "content": [
                {
                    "type":"image_url",
                    "image_url":{
                        "url":f"data:image/png;base64,{encode_image(obs['screenshot'])}",
                        "detail": "high"
                    },
                },
                {
                    "type": "text",
                    "text": user_prompt
                },
            ],
        })
        
        if self.planner_model == "o3":
            planner_response = self.call_llm(
                {
                    "model": self.planner_model,
                    "messages": messages,
                    "max_completion_tokens": self.max_tokens
                },
                self.planner_model,
            )
        else:
            planner_response = self.call_llm(
                {
                    "model": self.planner_model,
                    "messages": messages,
                    "max_tokens": self.max_tokens,
                    "top_p": self.top_p,
                    "temperature": self.temperature,
                },
                self.planner_model,
            )

        logger.info(f"Planner Output: {planner_response}")
        codes = self.parse_code_from_planner_response(planner_response)
        # Add retry logic if no codes were parsed
        retry_count = 0
        max_retries = 5
        while not codes and retry_count < max_retries:
            logger.info(f"No codes parsed from planner response. Retrying ({retry_count+1}/{max_retries})...")
            messages.append({
                "role": "user",
                "content": [
                    {"type": "text", "text": "You didn't generate valid actions. Please try again."}
                ]
            })
            if self.planner_model == "o3":
                planner_response = self.call_llm(
                    {
                        "model": self.planner_model,
                        "messages": messages,
                        "max_completion_tokens": self.max_tokens
                    },
                    self.planner_model,
                )
            else:
                planner_response = self.call_llm(
                    {
                        "model": self.planner_model,
                        "messages": messages,
                        "max_tokens": self.max_tokens,
                        "top_p": self.top_p,
                        "temperature": self.temperature,
                    },
                    self.planner_model,
                )
            logger.info(f"Retry Planner Output: {planner_response}")
            codes = self.parse_code_from_planner_response(planner_response)
            retry_count += 1
            
        thought = self.parse_thought_from_planner_response(planner_response)
        observation_caption = self.parse_observation_caption_from_planner_response(planner_response)
        resized_height, resized_width = smart_resize(height, width, max_pixels= 2700 * 28 * 28)
        pyautogui_actions = []
        for line in codes:
            code = self.convert_action_to_grounding_model_instruction(
                line,
                obs,
                instruction,
                height,
                width,
                resized_height,
                resized_width
            )
            pyautogui_actions.append(code)
        self.actions.append([pyautogui_actions])
        self.observations.append(obs)
        self.thoughts.append(thought)
        self.observation_captions.append(observation_caption)
        self.current_step += 1
        return planner_response, pyautogui_actions
        
    def parse_observation_caption_from_planner_response(self, input_string: str) -> str:
        pattern = r"Observation:\n(.*?)\n"
        matches = re.findall(pattern, input_string, re.DOTALL)
        if matches:
            return matches[0].strip()
        return ""

    def parse_thought_from_planner_response(self, input_string: str) -> str:
        pattern = r"Thought:\n(.*?)\n"
        matches = re.findall(pattern, input_string, re.DOTALL)
        if matches:
            return matches[0].strip()
        return ""

    def parse_code_from_planner_response(self, input_string: str) -> List[str]:

        input_string = "\n".join([line.strip() for line in input_string.split(';') if line.strip()])
        if input_string.strip() in ['WAIT', 'DONE', 'FAIL']:
            return [input_string.strip()]

        pattern = r"```(?:\w+\s+)?(.*?)```"
        matches = re.findall(pattern, input_string, re.DOTALL)
        codes = []

        for match in matches:
            match = match.strip()
            commands = ['WAIT', 'DONE', 'FAIL']

            if match in commands:
                codes.append(match.strip())
            elif match.split('\n')[-1] in commands:
                if len(match.split('\n')) > 1:
                    codes.append("\n".join(match.split('\n')[:-1]))
                codes.append(match.split('\n')[-1])
            else:
                codes.append(match)

        return codes

    def convert_action_to_grounding_model_instruction(self, line: str, obs: Dict, instruction: str, height: int, width: int, resized_height: int, resized_width: int ) -> str:
        pattern = r'(#.*?)\n(pyautogui\.(moveTo|click|rightClick|doubleClick|middleClick|dragTo)\((?:x=)?(\d+)(?:,\s*|\s*,\s*y=)(\d+)(?:,\s*duration=[\d.]+)?\))'
        matches = re.findall(pattern, line, re.DOTALL)
        if not matches:
            return line
        new_instruction = line
        for match in matches:
            comment = match[0].split("#")[1].strip()
            original_action = match[1]
            func_name = match[2].strip()

            if "click()" in original_action.lower():
                continue
            
            messages = []
            messages.append({
                "role": "system",
                "content": [{"type": "text", "text": JEDI_GROUNDER_SYS_PROMPT.replace("{height}", str(resized_height)).replace("{width}", str(resized_width))}]
            })
            messages.append(
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/png;base64,{encode_image(obs['screenshot'])}",
                                "detail": "high",
                            },
                        },
                        {
                            "type": "text",
                            "text": '\n' + comment,
                        },
                    ],
                }
            )
            grounding_response = self.call_llm({
                "model": self.executor_model,
                "messages": messages,
                "max_tokens": self.max_tokens,
                "top_p": self.top_p,
                "temperature": self.temperature
            }, self.executor_model)
            coordinates = self.parse_jedi_response(grounding_response, width, height, resized_width, resized_height)
            logger.info(coordinates)
            if coordinates == [-1, -1]:
                continue
            action_parts = original_action.split('(')
            new_action = f"{action_parts[0]}({coordinates[0]}, {coordinates[1]}"
            if len(action_parts) > 1 and 'duration' in action_parts[1]:
                duration_part = action_parts[1].split(',')[-1]
                new_action += f", {duration_part}"
            elif len(action_parts) > 1 and 'button' in action_parts[1]:
                button_part = action_parts[1].split(',')[-1]
                new_action += f", {button_part}"
            else:
                new_action += ")"
            logger.info(new_action)
            new_instruction = new_instruction.replace(original_action, new_action)
        return new_instruction
        
    def parse_jedi_response(self, response, width: int, height: int, resized_width: int, resized_height: int) -> List[str]:
        """
        Parse the LLM response and convert it to low level action and pyautogui code.
        """ 

        low_level_instruction = ""
        pyautogui_code = []
        try:
            # 定义可能的标签组合
            start_tags = ["<tool_call>", "⚗"]
            end_tags = ["</tool_call>", "⚗"]

            # 找到有效的开始和结束标签
            start_tag = next((tag for tag in start_tags if tag in response), None)
            end_tag = next((tag for tag in end_tags if tag in response), None)

            if not start_tag or not end_tag:
                print("The response is missing valid start or end tags")
                return low_level_instruction, pyautogui_code

            # 分割响应以提取low_level_instruction和tool_call
            parts = response.split(start_tag)
            if len(parts) < 2:
                print("The response is missing the start tag")
                return low_level_instruction, pyautogui_code

            low_level_instruction = parts[0].strip().replace("Action: ", "")
            tool_call_str = parts[1].split(end_tag)[0].strip()

            try:
                tool_call = json.loads(tool_call_str)
                action = tool_call.get("arguments", {}).get("action", "")
                args = tool_call.get("arguments", {})
            except json.JSONDecodeError as e:
                print(f"JSON parsing error: {e}")
                # 处理解析错误，返回默认值或空值
                action = ""
                args = {}
            
            # convert the coordinate to the original resolution
            x = int(args.get("coordinate", [-1, -1])[0] * width / resized_width)
            y = int(args.get("coordinate", [-1, -1])[1] * height / resized_height)

            return [x, y]
        except Exception as e:
            logger.error(f"Failed to parse response: {e}")
            return [-1, -1]

    @backoff.on_exception(
        backoff.constant,
        # here you should add more model exceptions as you want,
        # but you are forbidden to add "Exception", that is, a common type of exception
        # because we want to catch this kind of Exception in the outside to ensure
        # each example won't exceed the time limit
        (
            # General exceptions
            SSLError,
            # OpenAI exceptions
            openai.RateLimitError,
            openai.BadRequestError,
            openai.InternalServerError,
            # Google exceptions
            InvalidArgument,
            ResourceExhausted,
            InternalServerError,
            BadRequest,
            # Groq exceptions
            # todo: check
        ),
        interval=30,
        max_tries=10,
    )
    def call_llm(self, payload, model):
        if model.startswith("gpt") or model.startswith("o3"):
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {OPENAI_API_KEY}"
            }
            logger.info("Generating content with GPT model: %s", model)
            response = requests.post(
                "https://api.openai.com/v1/chat/completions",
                headers=headers,
                json=payload,
            )

            if response.status_code != 200:
                logger.error("Failed to call LLM: " + response.text)
                time.sleep(5)
                return ""
            else:
                return response.json()["choices"][0]["message"]["content"]

        elif model.startswith("jedi"):
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {JEDI_API_KEY}"
            }
            response = requests.post(
                f"{JEDI_SERVICE_URL}/v1/chat/completions",
                headers=headers,
                json=payload,
            )
            if response.status_code != 200:
                logger.error("Failed to call LLM: " + response.text)
                time.sleep(5)
                return ""
            else:
                return response.json()["choices"][0]["message"]["content"]

    def reset(self, _logger=None):
        global logger
        logger = (_logger if _logger is not None else
                  logging.getLogger("desktopenv.jedi_7b_agent"))

        self.thoughts = []
        self.action_descriptions = []
        self.actions = []
        self.observations = []
        self.observation_captions = []
